import os


from ..dataset import DatasetClient

api_host = "http://192.168.1.35:8002"
workspace_name = "default_workspace"
api_key = "4fc086fe-d64b-4ac8-816a-0dcbf34f7368"
workspace_id = "8fb7c66a-a49d-46a8-8044-51a94cee1a36"

DatasetClient.connect(api_host=api_host, api_key=api_key, workspace_name=workspace_name)

assert os.environ["OIP_API_HOST"] == api_host
assert os.environ["WORKSPACE_ID"] == workspace_id
assert os.environ["ML_DATASET_AUTHENTICATION_TOKEN"] == api_key

europe_dataset = DatasetClient.create(name="europe_dataset")
europe_dataset.add_files(
    "/Users/mohamed/projects/oip-mlops-be/client/oip_ml_dataset/examples/countries_dataset/europe"
)
europe_dataset.upload()
europe_dataset.finalize()

asia_dataset = DatasetClient.create(name="asia_dataset")
asia_dataset.add_files(
    "/Users/mohamed/projects/oip-mlops-be/client/oip_ml_dataset/examples/countries_dataset/asia",
    parent_datasets=[europe_dataset.id],
)
asia_dataset.upload()
asia_dataset.finalize()

africa_dataset = DatasetClient.create(name="africa_dataset")
africa_dataset.add_files(
    "/Users/mohamed/projects/oip-mlops-be/client/oip_ml_dataset/examples/countries_dataset/africa",
    parent_datasets=[asia_dataset.id],
)
africa_dataset.upload()
africa_dataset.finalize()

north_america_dataset = DatasetClient.create(name="north_america_dataset")
africa_dataset.add_files(
    "/Users/mohamed/projects/oip-mlops-be/client/oip_ml_dataset/examples/countries_dataset/north_america",
    parent_datasets=[africa_dataset.id],
)
north_america_dataset.upload()
north_america_dataset.finalize()



import os
import json
import boto3
from tqdm import tqdm
from typeguard import typechecked
from pathlib import Path
from .StorageProvider import StorageProvider
from concurrent.futures import ThreadPoolExecutor
from typing import Dict


class AwsProvider(StorageProvider):
    @staticmethod
    @typechecked
    def download(
        storage_conf_name: str,
        download_uri: str,
        download_folder: str,
        max_workers: int,
        verbose: bool,
    ) -> str:
        """
        Download all the dataset files located in a specific download_uri
        :param storage_conf_name: str: storage configuration name serve as an identifier.
        :param download_uri: str: uri from where we download the dataset files refers to the specific location or path
            within the storage provider where the dataset files are hosted.
            example s3://bucket/folder/../dataset_folder
        :param download_folder: str: destination directory or folder where downloaded files will be stored.
        :param max_workers: int:Numbers of threads to be spawned when zipping and uploading the files.
            If None number of logical cores
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: str: path to the downloaded dataset directory
        """
        aws_conf: Dict[str, str] = json.loads(os.environ[storage_conf_name])
        session: boto3.Session = boto3.Session(
            aws_access_key_id=aws_conf["ACCESS_KEY"],
            aws_secret_access_key=aws_conf["SECRET_KEY"],
            region_name=aws_conf["REGION"],
        )
        s3 = session.resource("s3")

        bucket_name: str = download_uri[len("s3://") :].split("/")[0]
        object_to_download: str = download_uri[len(f"s3://{bucket_name}/") :]
        dataset_name: str = download_uri.split("/")[-2]
        bucket = s3.Bucket(bucket_name)
        bucket_objects = bucket.objects.filter(Prefix=object_to_download)

        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            if verbose:
                download_prog_bar = tqdm(
                    total=len(list(bucket_objects)),
                    desc="downloading",
                    unit=" file ",
                )
            for obj in bucket_objects:
                file_path: Path = Path(obj.key)
                parent_folder: Path = Path(f"{download_folder}/{str(file_path.parent)}")
                if not parent_folder.exists():
                    parent_folder.mkdir(parents=True, exist_ok=True)
                download_task = pool.submit(
                    bucket.download_file,
                    str(file_path),
                    f"{download_folder}/{str(file_path)}",
                )
                if verbose:
                    download_task.add_done_callback(
                        lambda task: download_prog_bar.update(1)
                    )

        return f"{download_folder}/{object_to_download}"



from typing import Dict, List, Optional, TypedDict, Any
from enum import Enum

# entity instance
EntityInstance = Dict[str, Any]

# list of entities instances
EntityListInstances = List[EntityInstance]


# class ApiRequestArgs(TypedDict, total=False):
#     filter_cols: Optional[str]
#     filter_ops: Optional[str]
#     filter_vals: Optional[str]
#     page_num: Optional[str]
#     page_size: Optional[str]
#     sort_col: Optional[str]
#     sort_order: Optional[str]
#     logical_op: Optional[str]

ApiRequestArgs = Dict[str, str]


class FileEntryDict(TypedDict, total=False):
    relative_path: str
    parent_dataset_id: str
    artifact_name: Optional[str]
    local_path: Optional[str]
    size: Optional[int]
    hash: Optional[str]
    remote_url_basename: Optional[str]


FileEntries = Dict[str, FileEntryDict]


class DatasetDict(TypedDict, total=False):
    name: str
    id: str
    workspace_id: str
    version: str
    is_main: bool
    parent_datasets: List[str]
    _tags: List[str]
    status: str
    file_entries: List[FileEntryDict]
    file_count: int
    added_files: int
    modified_files: int
    removed_files: int
    total_size: int
    total_size_compressed: int
    local_download_folder: str
    description: Optional[str]
    dependency_graph: Dict[str, List[str]]
    dependency_chunk_lookup: Dict[str, int]


class DatasetState(TypedDict, total=False):
    id: str
    file_entries: List[FileEntryDict]
    file_count: int
    added_files: int
    modified_files: int
    removed_files: int
    total_size: int
    total_size_compressed: int
    dependency_graph: Dict[str, List[str]]
    dependency_chunk_lookup: Optional[Dict[str, int]]


class DatasetStatus(Enum):
    CREATED: str = "created"
    IN_PROGRESS: str = "in_progress"
    UPLOADED: str = "uploaded"
    FAILED: str = "failed"
    COMPLETED: str = "completed"
    ABORTED: str = "aborted"
    UNKNOWN: str = "unknown"



import os
from ..dataset import DatasetClient

api_host = "http://192.168.1.35:8002"
workspace_name = "default_workspace"
api_key = "4fc086fe-d64b-4ac8-816a-0dcbf34f7368"
workspace_id = "8fb7c66a-a49d-46a8-8044-51a94cee1a36"

DatasetClient.connect(api_host=api_host, api_key=api_key, workspace_name=workspace_name)

assert os.environ["OIP_API_HOST"] == api_host
assert os.environ["WORKSPACE_ID"] == workspace_id
assert os.environ["ML_DATASET_AUTHENTICATION_TOKEN"] == api_key

dataset = DatasetClient.create(name="my_dataset")
dataset.add_files(
    "/Users/mohamed/projects/oip-mlops-be/client/oip_ml_dataset/examples/countries_dataset/europe"
)
dataset.remove_files("Germany.png")
dataset.upload()
dataset.finalize()



import os
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Optional, List, Dict, Set, Deque
from zipfile import ZipFile, ZIP_DEFLATED
from six.moves.queue import PriorityQueue, Queue, Empty
from pathlib import Path
from tempfile import mkstemp
from collections import deque
from tqdm import tqdm


class ParallelZipper(object):
    """
    Used to zip multiple files in zip chunks of a particular size, all in parallel
    """

    class ZipperObject(object):
        def __init__(
            self,
            chunk_size: int,  # int
            zipper_queue: PriorityQueue,
            zipper_results: Queue,
            allow_zip_64: bool,
            compression: Any,
            zip_prefix: str,
            zip_suffix: str,
        ):
            # (...) -> ParallelZipper.ZipperObject
            """
            Initialize a ParallelZipper.ZipperObject instance that holds its corresponding zip
            file, as well as other relevant data

            :param chunk_size: Chunk size, in MB. The ParallelZipper will try its best not to exceed this size
                when bulding up this zipper object, but that is not guaranteed
            :param zipper_queue: PriorityQueue that holds ParallelZipper.ZipperObject instances.
                When this ParallelZipper.ZipperObject can hold more data (i.e. chunk_size was not exceeded),
                this object will reinsert itself in this queue to be reused by the ParallelZipper.
                Else, a fresh ParallelZipper.ZipperObject will be inserted
            :param zipper_results: Queue that holds ParallelZipper.ZipperObject instances. These instances
                are added to this queue when chunk_size is exceeded
            :param allow_zip_64: if True, ZipFile will create files with ZIP64 extensions when
                needed, otherwise it will raise an exception when this would be necessary
            :param compression: ZipFile.ZIP_STORED (no compression), ZipFile.ZIP_DEFLATED (requires zlib),
                ZipFile.ZIP_BZIP2 (requires bz2) or ZipFile.ZIP_LZMA (requires lzma).
            :param zip_prefix: The zip file created by this object will have its name prefixed by this
            :param zip_suffix: The zip file created by this object will have its name suffixed by this
            :return: ParallelZipper.ZipperObject instance
            """
            self._chunk_size: int = chunk_size
            self._zipper_queue = zipper_queue
            self._zipper_results = zipper_results
            self._allow_zip_64: bool = allow_zip_64
            self._compression: Any = compression
            self._zip_prefix: str = zip_prefix
            self._zip_suffix: str = zip_suffix
            self.fd, zip_path = mkstemp(prefix=zip_prefix, suffix=zip_suffix)
            self.zip_path = Path(zip_path)
            self.zip_file = ZipFile(
                self.zip_path.as_posix(),
                "w",
                allowZip64=allow_zip_64,
                compression=compression,
            )
            self.archive_preview: List[Dict[str, str]] = []
            self.count: int = 0
            self.files_zipped: Set = set()

        def zip(self, file_path: str, zip_name: str) -> None:
            """
            Zips a file into the ZipFile created by this instance. This instance will either add
            itself back to the PriorityQueue used to select the best zipping candidate or add itself
            to the result Queue after exceeding self.chunk_size.

            :param file_path: str: Path to the file to be zipped
            :param zip_name: str: Name of the file in the archive
            """
            self.zip_file.write(file_path, zip_name)
            self.count += 1
            self.files_zipped.add(Path(file_path).as_posix())
            if self._chunk_size <= 0 or self.size < self._chunk_size:
                self._zipper_queue.put(self)
            else:
                self._zipper_queue.put(
                    ParallelZipper.ZipperObject(
                        self._chunk_size,
                        self._zipper_queue,
                        self._zipper_results,
                        self._allow_zip_64,
                        self._compression,
                        self._zip_prefix,
                        self._zip_suffix,
                    )
                )
                self._zipper_results.put(self)

        def merge(self, other) -> None:
            """
            Merges one ParallelZipper.ZipperObject instance into the current one.
            All the files zipped by the other instance will be added to this instance,
            as well as any other useful additional data

            :param other: ParallelZipper.ZipperObject: ParallelZipper.ZipperObject instance to merge into this one
            """
            with ZipFile(self.zip_path.as_posix(), "a") as parent_zip:
                with ZipFile(other.zip_path.as_posix(), "r") as child_zip:
                    for child_name in child_zip.namelist():
                        parent_zip.writestr(
                            child_name, child_zip.open(child_name).read()
                        )
            self.files_zipped |= other.files_zipped
            self.count += other.count

        def close(self) -> None:
            """
            Attempts to close file descriptors associated to the ZipFile
            """
            try:
                self.zip_file.close()
                os.close(self.fd)
            except (Exception,):
                pass

        def delete(self) -> None:
            """
            Attempts to delete the ZipFile from the disk
            """
            try:
                self.close()
                self.zip_path.unlink()
            except (Exception,):
                pass

        @property
        def size(self) -> int:
            """
            :return: Size of the ZipFile, in bytes
            """
            return self.zip_path.stat().st_size

        def __lt__(self, other):
            # we want to completely "fill" as many zip files as possible, hence the ">" comparison
            return self.size > other.size

    def __init__(
        self,
        chunk_size: int,
        max_workers: int,
        allow_zip_64: Optional[bool] = True,
        compression: Optional[Any] = ZIP_DEFLATED,
        zip_prefix: Optional[str] = "",
        zip_suffix: Optional[str] = "",
        pool: Optional[ThreadPoolExecutor] = None,
    ) -> None:
        """
        Initialize the ParallelZipper. Each zip created by this object will have the following naming
        format: [zip_prefix]<random_string>[zip_suffix]

        :param chunk_size: Chunk size, in MB. The ParallelZipper will try its best not to exceed this size,
            but that is not guaranteed
        :param max_workers: The maximum number of workers spawned when zipping the files
        :param allow_zip_64: if True, ZipFile will create files with ZIP64 extensions when
            needed, otherwise it will raise an exception when this would be necessary
        :param compression: ZipFile.ZIP_STORED (no compression), ZipFile.ZIP_DEFLATED (requires zlib),
            ZipFile.ZIP_BZIP2 (requires bz2) or ZipFile.ZIP_LZMA (requires lzma).
        :param zip_prefix: Zip file names will be prefixed by this
        :param zip_suffix: Zip file names will pe suffixed by this
        :param pool: Use this ThreadPoolExecutor instead of creating one. Note that this pool will not be
            closed after zipping is finished.

        :return: ParallelZipper instance
        """
        # chunk_size in Megabyte
        self._chunk_size: int = chunk_size * (1024**2)
        self._max_workers: int = max_workers
        self._allow_zip_64: bool = allow_zip_64
        self._compression: Any = compression
        self._zip_prefix: str = zip_prefix
        self._zip_suffix: str = zip_suffix
        self._pool: ThreadPoolExecutor = pool
        self._zipper_queue: PriorityQueue = PriorityQueue()
        self._zipper_results: Queue = Queue()

    def zip_iter(
        self,
        file_paths: List[str],
        zip_names: Dict[str, str],
        total_chunks: int,
        verbose: Optional[bool] = True,
    ):
        """
        Generator function that returns zip files as soon as they are available.
        The zipping is done in parallel

        :param file_paths: List[str]: List of paths to the files to zip
        :param zip_names: dict: example {abs_local_path_to_file_1:relative_path_to_file_1,abs_local_path_to_file_2:..}
        :param total_chunks: int: Expected number of chunks to zip
            (total_chunks value is an approximate estimation it may be lower)
        :return: Generator of ParallelZipper.ZipperObjects
        """
        while not self._zipper_queue.empty():
            self._zipper_queue.get_nowait()
        for _ in range(self._max_workers):
            self._zipper_queue.put(
                ParallelZipper.ZipperObject(
                    self._chunk_size,
                    self._zipper_queue,
                    self._zipper_results,
                    self._allow_zip_64,
                    self._compression,
                    self._zip_prefix,
                    self._zip_suffix,
                )
            )
        for file_path in file_paths[:]:
            if not Path(file_path).is_file():
                file_paths.remove(file_path)

        file_paths = sorted(
            file_paths, key=lambda k: Path(k).stat().st_size, reverse=True
        )
        # zip in parallel
        pooled: List = []
        if verbose:
            zip_pbar: tqdm = tqdm(total=total_chunks, desc="zipping", unit=" chunk")
        if not self._pool:
            pool: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=self._max_workers)
        else:
            pool = self._pool
        for f in file_paths:
            zipper = self._zipper_queue.get()
            pooled.append(pool.submit(zipper.zip, Path(f).as_posix(), zip_names.get(f)))
            # yield zipped files
            for result in self._yield_zipper_results():
                yield result
                if verbose:
                    zip_pbar.update(1)
        # await task completion
        for task in pooled:
            task.result()
        #  if no active tasks in the pool
        if not self._pool:
            pool.shutdown()

        for result in self._yield_zipper_results():
            yield result
        # files that have been compressed but are still in the queue
        zipper_results_leftover: Deque = deque()

        # extract remaining results
        while not self._zipper_queue.empty():
            result = self._zipper_queue.get()
            if result.count != 0:
                zipper_results_leftover.append(result)
            else:
                result.delete()
        # double-ended queue.
        # It is a data structure that allows you to efficiently add and remove elements from both ends
        zipper_results_leftover = deque(sorted(zipper_results_leftover, reverse=True))

        # merge zip files greedily if possible and get the paths as results
        while len(zipper_results_leftover) > 0:
            zip_ = zipper_results_leftover.pop()
            zip_.close()
            if zip_.size >= self._chunk_size > 0:
                yield zip_
                if verbose:
                    zip_pbar.update(1)
                continue
            while len(zipper_results_leftover) > 0 and (
                self._chunk_size <= 0
                or zipper_results_leftover[0].size + zip_.size < self._chunk_size
            ):
                child_zip = zipper_results_leftover.popleft()
                child_zip.close()
                zip_.merge(child_zip)
                child_zip.delete()
            yield zip_
            if verbose:
                zip_pbar.update(1)
        if verbose:
            zip_pbar.total = zip_pbar.n

    def _yield_zipper_results(self):
        while True:
            try:
                result = self._zipper_results.get_nowait()
                result.close()
                yield result
            except Empty:
                break



from copy import deepcopy
from attr import attrs, attrib

import re
import six

if six.PY3:
    from math import inf
else:
    inf = float("inf")
from typing import Tuple, List, Union


class InvalidVersion(ValueError):
    """
    An invalid version was found, users should refer to PEP 440.
    """


@attrs
class _Version(object):
    """
    1- epoch: A numeric value indicating major version transitions. For example,
        if a software project has a major structural change, the epoch value would increment (e.g., from 1 to 2).
    2- release: Represents the stable version of the software. For instance,
        if the software is in a production-ready state, it would have a release version (e.g., "2.0.1").
    3- dev: Marks versions in active development, typically indicated by appending "dev" and a
        number to the version (e.g., "2.1.0dev1" for the first development version of the upcoming release).
    4- pre: Denotes pre-release versions that are available before the stable release,
        often marked with an "a" for alpha, "b" for beta, or "rc" for release candidate
        (e.g., "2.1.0a1" for the first alpha version).
    5- post: Identifies post-release updates, usually with a "post" and a number
        (e.g., "2.1.0post1" for the first post-release update).
    6- local: Contains local-specific version information, allowing developers to add custom version
        details for their specific needs. For example, a local version may look like "2.1.0+myappbuild42.
    """

    epoch = attrib()
    release = attrib()
    dev = attrib()
    pre = attrib()
    post = attrib()
    local = attrib()


class _BaseVersion(object):
    def __init__(self, key):
        self._key = key

    def __hash__(self):
        return hash(self._key)

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)

    def _compare(self, other, method):
        if not isinstance(other, _BaseVersion):
            return NotImplemented

        return method(self._key, other._key)


class Version(_BaseVersion):
    VERSION_PATTERN = r"""
        v?
        (?:
            (?:(?P<epoch>[0-9]+)!)?                           # epoch
            (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
            (?P<pre>                                          # pre-release
                [-_\.]?
                (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
                [-_\.]?
                (?P<pre_n>[0-9]+)?
            )?
            (?P<post>                                         # post release
                (?:-(?P<post_n1>[0-9]+))
                |
                (?:
                    [-_\.]?
                    (?P<post_l>post|rev|r)
                    [-_\.]?
                    (?P<post_n2>[0-9]+)?
                )
            )?
            (?P<dev>                                          # dev release
                [-_\.]?
                (?P<dev_l>dev)
                [-_\.]?
                (?P<dev_n>[0-9]+)?
            )?
        )
        (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
    """
    _regex = re.compile(r"^\s*" + VERSION_PATTERN + r"\s*$", re.VERBOSE | re.IGNORECASE)
    _local_version_separators = re.compile(r"[\._-]")

    def __init__(self, version):
        # Validate the version and parse it into pieces
        match = self._regex.search(version)
        if not match:
            raise InvalidVersion(f"Invalid version: '{version}'")
        # We consider there to be an implicit 0 in a (pre or post or dev)-release if there is
        # not a numeral associated with it.
        pre_number: int = 0
        if match.group("pre_n"):
            pre_number = int(match.group("pre_n"))
        post_number: int = 0
        if match.group("post_n1"):
            post_number = int(match.group("post_n1"))
        elif match.group("post_n2"):
            post_number = int(match.group("post_n2"))
        dev_number: int = 0
        if match.group("dev_n"):
            dev_number = int(match.group("dev_n"))

        # Store the parsed out pieces of the version
        self._version = _Version(
            epoch=int(match.group("epoch")) if match.group("epoch") else 0,
            release=tuple(int(i) for i in match.group("release").split(".")),
            pre=self._parse_letter_version(match.group("pre_l"), pre_number),
            post=self._parse_letter_version(
                match.group("post_l") or "",
                post_number,
            ),
            dev=self._parse_letter_version(match.group("dev_l") or "", dev_number),
            local=self._parse_local_version(match.group("local") or ""),
        )

        # Generate a key which will be used for sorting
        key = self._cmpkey(
            self._version.epoch,
            self._version.release,
            self._version.pre,
            self._version.post,
            self._version.dev,
            self._version.local,
        )

        super(Version, self).__init__(key)

    def __repr__(self) -> str:
        return f"<Version({repr(str(self))})>"

    def __str__(self) -> str:
        parts: List[str] = []

        # Epoch
        if self.epoch != 0:
            parts.append(f"{self.epoch}!")

        # Release segment
        parts.append(".".join(str(x) for x in self.release))

        # Pre-release
        if self.pre is not None:
            parts.append("".join(str(x) for x in self.pre))

        # Post-release
        if self.post is not None:
            parts.append(f".post{self.post}")

        # Development release
        if self.dev is not None:
            parts.append(f".dev{self.dev}")

        # Local version segment
        if self.local is not None:
            parts.append(f"+{self.local}")

        return "".join(parts)

    def get_next_version(self):
        def increment(part: Union[int, List]):
            """
            :param part:union[int,list]:
            :return:  the object's type
            """
            if isinstance(part, int):
                return part + 1
            type_ = type(part)
            part = list(part)
            if isinstance(part[-1], int):
                part[-1] += 1
            return type_(part)

        next_version = deepcopy(self)
        if next_version._version.dev:
            next_version._version.dev = increment(next_version._version.dev)
        elif next_version._version.post:
            next_version._version.post = increment(next_version._version.post)
        elif next_version._version.pre:
            next_version._version.pre = increment(next_version._version.pre)
        elif next_version._version.release:
            next_version._version.release = increment(next_version._version.release)
        elif next_version._version.epoch:
            next_version._version.epoch = increment(next_version._version.epoch)
        return next_version

    @property
    def epoch(self):
        return self._version.epoch

    @property
    def release(self):
        return self._version.release

    @property
    def pre(self):
        return self._version.pre

    @property
    def post(self):
        return self._version.post[1] if self._version.post else None

    @property
    def dev(self):
        return self._version.dev[1] if self._version.dev else None

    @property
    def local(self):
        if self._version.local:
            return ".".join(str(x) for x in self._version.local)
        else:
            return None

    @property
    def public(self):
        return str(self).split("+", 1)[0]

    @property
    def base_version(self) -> str:
        parts: List[str] = []

        # Epoch
        if self.epoch != 0:
            parts.append(f"{self.epoch}!")

        # Release segment
        parts.append(".".join(str(x) for x in self.release))

        return "".join(parts)

    @property
    def is_prerelease(self):
        return self.dev is not None or self.pre is not None

    @property
    def is_postrelease(self):
        return self.post is not None

    @property
    def is_devrelease(self):
        return self.dev is not None

    @staticmethod
    def _parse_letter_version(letter: str, number: int) -> Union[Tuple[str, int], None]:
        """
        Parse letter Version
        :param letter: str: latter examples: alpha,beta....
        :param number: int:
        :return: Union[Tuple[str, int], None]: return (letter,version) or none
        """
        if not letter and not number:
            return None
        if letter:
            # We normalize any letters to their lower case form
            letter = letter.lower()

            # We consider some words to be alternate spellings of other words and
            # in those cases we want to normalize the spellings to our preferred
            # spelling.
            if letter == "alpha":
                letter = "a"
            elif letter == "beta":
                letter = "b"
            elif letter in ["c", "pre", "preview"]:
                letter = "rc"
            elif letter in ["rev", "r"]:
                letter = "post"

            return letter, int(number)
        if not letter and number:
            # We assume if we are given a number, but we are not given a letter
            # then this is using the implicit post release syntax (e.g. 1.0-1)
            letter = "post"

        return letter, int(number)

    @classmethod
    def is_valid_version_string(cls, version_string: str) -> bool:
        """
        Check iv the version is a valide one
        :param version_string: str: version we want to check
        :return: bool: True if the version is valid

        """
        if not version_string:
            return False
        match = cls._regex.search(version_string)
        return bool(match)

    @classmethod
    def _parse_local_version(cls, local: str) -> Union[Tuple, None]:
        """
        Takes a string like abc.1.twelve and turns it into ("abc", 1, "twelve").
        :param local: str: local segment
        :return: union[tuple,non]
        """
        if local is not None:
            local_ = tuple(
                part.lower() if not part.isdigit() else int(part)
                for part in cls._local_version_separators.split(local)
            )
            if not local_ or not local_[0]:
                return None
            return local_
        return None

    @staticmethod
    def _cmpkey(epoch, release, pre, post, dev, local):
        """
        Generate a key which will be used for sorting
        :param epoch:int:  A numeric value indicating major version transitions (e.g., from 1 to 2)
        :param release: tuple(int,int,int): Represents the stable version of the software (e.g (2,0,1))
        :param pre: tuple(str, int): Denotes pre-release versions that are available before the stable release,
            often marked with an "a" for alpha,
            "b" for beta, or "rc" for release candidate (e.g., pre:(a,1) "2.1.0a1"  for the first alpha version).
        :param post: tuple(str, int) : Identifies post-release updates, usually with a "post" and a number
            (e.g., post:(post,1) "2.1.0post1" for the first post-release update).
        :param dev: tuple(str, int):Marks versions in active development (e.g., dev: (dev,1) "2.1.0dev1")
        :param local: local-specific version information, allowing developers to add custom version
        :return: tuple
        """
        # When we compare a release version, we want to compare it with all of the
        # trailing zeros removed. So we'll use a reverse the list, drop all the now
        # leading zeros until we come to something non zero, then take the rest
        # re-reverse it back into the correct order and make it a tuple and use
        # that for our sorting key.
        # release = tuple(
        #     reversed(list(itertools.dropwhile(lambda x: x == 0, reversed(release))))
        # )

        # Versions without a pre-release (except as noted above) should sort after
        # those with one.
        if not pre:
            pre = inf
        elif pre:
            pre = pre[1]

        # Versions without a post segment should sort before those with one.
        if not post:
            post = -inf
        else:
            post = post[1]

        # Versions without a development segment should sort after those with one.
        if not dev:
            dev = inf
        else:
            dev = dev[1]

        if not local:
            # Versions without a local segment should sort before those with one.
            local = inf
        else:
            # Versions with a local segment need that segment parsed to implement
            # the sorting rules in PEP440.
            # - Alpha numeric segments sort before numeric segments
            # - Alpha numeric segments sort lexicographically
            # - Numeric segments sort numerically
            # - Shorter versions sort before longer versions when the prefixes
            #   match exactly
            local = local[1]

        return epoch, release, pre, post, dev, local



from .StorageProvider import StorageProvider
from typeguard import typechecked


class MinioProvider(StorageProvider):
    @staticmethod
    @typechecked
    def download(
        storage_conf_name: str,
        download_uri: str,
        download_folder: str,
        max_workers: int,
        verbose: bool,
    ) -> str:
        """
        Download all the dataset files located in a specific download_uri
        :param storage_conf_name: str: storage configuration name serve as an identifier.
        :param download_uri: str: uri from where we download the dataset files refers to the specific location or path
            within the storage provider where the dataset files are hosted.
            example minio://bucket/folder/../dataset_folder
        :param download_folder: str: destination directory or folder where downloaded files will be stored.
        :param max_workers: int:Numbers of threads to be spawned when zipping and uploading the files.
            If None number of logical cores
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: str: path to the downloaded dataset directory
        """
        raise NotImplementedError



import os
import hashlib
from functools import wraps
import json
from fnmatch import fnmatch
from dotenv import load_dotenv
from typing import Union, Optional, List, Tuple, Dict
from .logger import Logger
from typeguard import typechecked


logger = Logger().get_logger()


def check_connection(func):
    """
    Decorator to check if required environment variables are set.
    If not set, attempt to load them from .env file.
    If still not set, raise an exception.
    """

    REQUIRED_ENV_VARS = [
        "OIP_API_HOST",
        "OIP_API_KEY",
        "OIP_WORKSPACE_NAME",
        "OIP_WORKSPACE_ID",
    ]

    def wrapper(*args, **kwargs):
        # Check if all required environment variables are set
        if all(var in os.environ for var in REQUIRED_ENV_VARS):
            return func(*args, **kwargs)

        # Attempt to load environment variables from .oip_dataset_env file
        load_dotenv(".oip_dataset_env")

        # Check again if all required environment variables are set
        if all(var in os.environ for var in REQUIRED_ENV_VARS):
            return func(*args, **kwargs)
        else:
            raise Exception(
                """Not connected, you must establish a connection either through the SDK using the connect function 
                or via the command line interface (CLI) by running oip-dataset-client-init."""
            )

    return wrapper


@typechecked
def check_storage_conf(storage_conf_name: str):
    """
    check if the storage configuration exist
    :param storage_conf_name: str: storage configuration name
    :return: none
    """
    if storage_conf_name not in os.environ:
        raise ValueError(
            f"Config {storage_conf_name} doesn't exist please add it before starting your migration."
        )
    # load storage configuration
    storage_conf: Dict[str, str] = json.loads(os.environ[storage_conf_name])

    if "PROVIDER" not in storage_conf:
        raise KeyError(
            "Provider information missing. Please review and re-add your configuration."
        )
    provider: str = storage_conf["PROVIDER"]

    providers_config_keys: Dict[str, List[str]] = {
        "AWS": ["ACCESS_KEY", "SECRET_KEY", "REGION"],
        "MINIO": ["HOST", "ACCESS_KEY", "SECRET_KEY"],
        "GCP": ["CREDENTIALS_JSON", "PROJECT", "ACCOUNT_NAME", "BUCKET"],
        "AZURE": ["ACCOUNT_NAME", "ACCOUNT_KEY", "BUCKET"],
    }

    if provider in providers_config_keys and not any(
        key in storage_conf for key in providers_config_keys[provider]
    ):
        raise Exception(
            f"Missing config for {storage_conf_name}. Please review and re-add your configuration."
        )
    elif provider not in providers_config_keys and provider != "OTHER":
        raise Exception(
            f"Provider {provider} doesn't exist. Please review and re-add your configuration."
        )


def sha256sum(
    filename: str, skip_header: Optional[int] = 0, block_size: Optional[int] = 65536
) -> Union[Tuple[str, str], Tuple[None, None]]:
    """
    create sha2 of the file, notice we skip the header of the file (32 bytes)
        because sometimes that is the only change
    :param filename: str: file name
    :param skip_header: int: default to 0 if 1 we skip
    :param block_size: int: default to 65536
    :return: union[(str,str),(none,none)]:
    """
    h = hashlib.sha256()
    file_hash = hashlib.sha256()
    b = bytearray(block_size)
    mv = memoryview(b)
    try:
        with open(filename, "rb", buffering=0) as f:
            # skip header
            if skip_header:
                file_hash.update(f.read(skip_header))
            for n in iter(lambda: f.readinto(mv), 0):
                h.update(mv[:n])
                if skip_header:
                    file_hash.update(mv[:n])
    except (Exception,):
        return None, None

    return h.hexdigest(), file_hash.hexdigest() if skip_header else None


def matches_any_wildcard(
    path: str, wildcards: Union[str, List], recursive: Optional[bool] = True
) -> bool:
    """
    Checks if given pattern matches any supplied wildcard

    :param path:str: path to check
    :param wildcards:union[str,list] wildcards to check against
    :param recursive: bool: whether or not the check is recursive. Default: True
        E.g. for path='directory/file.ext' and wildcards='*.ext',
        recursive=False will return False, but recursive=True will
        return True

    :return: bool: True if the path matches any wildcard and False otherwise
    """
    if wildcards is None:
        wildcards = ["*"]
    if not isinstance(wildcards, list):
        wildcards = [wildcards]
    wildcards = [str(w) for w in wildcards]
    if not recursive:
        path_as_list: List[str] = path.split("/")
    for wildcard in wildcards:
        if not recursive:
            wildcard = wildcard.split("/")
            matched: bool = True
            if len(path_as_list) != len(wildcard):
                continue
            for path_segment, wildcard_segment in zip(path_as_list, wildcard):
                if not fnmatch(path_segment, wildcard_segment):
                    matched = False
                    break
            if matched:
                return True
        else:
            wildcard_file: str = wildcard.split("/")[-1]
            wildcard_dir: str = wildcard[: -len(wildcard_file)] + "*"
            if fnmatch(path, wildcard_dir) and fnmatch(
                "/" + path, "*/" + wildcard_file
            ):
                return True
    return False


def format_size(
    size_in_bytes: Union[int, float],
    binary: Optional[bool] = False,
    use_nonbinary_notation: Optional[bool] = False,
    use_b_instead_of_bytes: Optional[bool] = False,
) -> str:
    """
    Return the size in human readable format (string)
    Matching humanfriendly.format_size outputs

    :param size_in_bytes:union[float,int]: number of bytes
    :param binary: bool: If `True` 1 Kb equals 1024 bytes, if False (default) 1 KB = 1000 bytes
    :param use_nonbinary_notation: bool: Only applies if binary is `True`. If this is `True`,
        the binary scale (KiB, MiB etc.) will be replaced with the regular scale (KB, MB etc.)
    :param use_b_instead_of_bytes: bool: If `True`, return the formatted size with `B` as the
        scale instead of `byte(s)` (when applicable)
    :return: string representation of the number of bytes (b,Kb,Mb,Gb, Tb,)
        >>> format_size(0)
        '0 bytes'
        >>> format_size(1)
        '1 byte'
        >>> format_size(5)
        '5 bytes'
        > format_size(1000)
        '1 KB'
        > format_size(1024, binary=True)
        '1 KiB'
        >>> format_size(1000 ** 3 * 4)
        '4 GB'
    """
    size: float = float(size_in_bytes)
    # single byte is the exception here
    if size == 1 and not use_b_instead_of_bytes:
        return "{} byte".format(int(size))
    k: int = 1024 if binary else 1000
    scale: List[str] = (
        ["bytes", "KiB", "MiB", "GiB", "TiB", "PiB"]
        if (binary and not use_nonbinary_notation)
        else ["bytes", "KB", "MB", "GB", "TB", "PB"]
    )
    if use_b_instead_of_bytes:
        scale[0] = "B"
    for i, m in enumerate(scale):
        if size < k ** (i + 1) or i == len(scale) - 1:
            return (
                (
                    "{:.2f}".format(size / (k**i)).rstrip("0").rstrip(".")
                    if i > 0
                    else "{}".format(int(size))
                )
                + " "
                + m
            )
    # we should never get here
    return f"{int(size)} {scale[0]}"


def is_within_directory(directory: str, target: str) -> bool:
    """
    Check if the path represented by 'target' is within or equal to the 'directory'.
    :param directory: str: The directory path to check against.
    :param target: str: The path to be checked.
    :return: bool: True if 'target' is within or equal to 'directory', False otherwise.
    """
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)
    prefix = os.path.commonprefix([abs_directory, abs_target])
    return prefix == abs_directory



""" Open Innovation Dataset Client configuration wizard"""

import os
from typing import Dict, Optional
from oip_dataset_client.HttpClient import HttpClient
from oip_dataset_client.schema import ApiRequestArgs
from oip_dataset_client.ServerConf import SERVER_MODULE_NAME, WORKSPACE_ENTITY_NAME
from oip_dataset_client.logger import Logger

logger = Logger().get_logger()


def main():
    """setting up the oip-dataset-client"""

    print("Open Innovation Dataset SDK setup process")
    oip_credentials = prompt_oip_credentials()
    with open(".oip_dataset_env", "w") as env_file:

        if oip_credentials:
            env_file.write(f"OIP_API_HOST={oip_credentials['api_host']}{os.linesep}")
            env_file.write(f"OIP_API_KEY={oip_credentials['api_key']}{os.linesep}")
            env_file.write(
                f"OIP_WORKSPACE_ID={oip_credentials['workspace_id']}{os.linesep}"
            )
            env_file.write(
                f"OIP_WORKSPACE_NAME={oip_credentials['workspace_name']}{os.linesep}"
            )
            logger.info("Connection to the server successful. Welcome!")


def prompt_oip_credentials() -> Optional[Dict[str, str]]:
    """
    :return: dict: open innovation credentials example:
        {api_host:oip_host,
        api_key:your_api_key,
        workspace_name:your_workspace_name,
        workspace_id:your_workspace_id}
    """
    description = (
        "\n"
        "Please navigate to the Workspace page within the `Open Innovation web application`.\n"
        "Select your desired workspace, then click on `Create new credentials`.\n"
        "Afterward, click on `Copy to clipboard.` Finally, paste the copied configuration here:\n"
    )
    print(description)
    # parse input to dict
    try:
        oip_credentials: Dict[str, str] = {}
        # Iterate over input until encountering "}"
        while True:
            line = input().strip()
            if line == "}":
                break
            if "=" in line:
                key, value = map(str.strip, line.split("="))
                value = value.replace('"', "")
                oip_credentials[key] = value
    except:
        raise Exception(
            "You need to copy and paste the correct configuration exactly as it appears on the website"
        )

    api_host: str = oip_credentials.get("api_host", "")
    api_key: str = oip_credentials.get("api_key", "")
    workspace_name: str = oip_credentials.get("workspace", "")
    # print message to the consol and exist if any issue happen
    workspace_id: Optional[str] = get_workspace_id(api_host, api_key, workspace_name)
    if workspace_id:
        oip_credentials["workspace_id"] = workspace_id
        return oip_credentials
    else:
        return None


def get_workspace_id(api_host: str, api_key: str, workspace_name: str) -> Optional[str]:
    """
    Retrieve workspace id from the server api based on the provided parameters.
    :param api_host: str: The API Host
    :param api_key: str: The API Key
    :param workspace_name: str: The workspace name
    :return: str: Workspace ID.
    """
    url: str = os.path.join(api_host, SERVER_MODULE_NAME, WORKSPACE_ENTITY_NAME)
    url = url.replace("\\", "/")
    headers: Dict[str, str] = {"authorization": "APIKey " + api_key}
    api_request_args: ApiRequestArgs = {
        "filter_cols": "name",
        "filter_ops": "=",
        "filter_vals": workspace_name,
    }
    response = HttpClient.get(url, headers, api_request_args)
    if response.status_code == 200 and response.json()["data"]:
        return response.json()["data"][0]["id"]
    else:
        print(
            f"{workspace_name} is not a valid workspace please verify it and try again"
        )
        exit(1)


if __name__ == "__main__":
    main()



import os
from typing import Dict, Optional, Union
from .HttpClient import HttpClient
from .schema import EntityListInstances, DatasetDict, ApiRequestArgs
from typeguard import typechecked
from .logger import Logger
from .ServerConf import SERVER_MODULE_NAME


logger = Logger().get_logger()


class ServerFacade:
    # the module that processes client requests on the server side.
    __server_module_name: str = SERVER_MODULE_NAME

    @classmethod
    @typechecked
    def save_data(
        cls,
        method: str,
        data: DatasetDict,
        dataset_entity_name: str,
    ):
        """
        Save the data on the server
        :param method: str: "Create" a new record if it doesn't exist, or "update" an existing record.
        :param data: DatasetDict: Data we need to create or update.
        :param dataset_entity_name: str: dataset entity name.
        :return: response content
        """
        logger.debug("call save_data")
        if method not in ["create", "update"]:
            raise ValueError(f"METHOD {method} IS NOT ALLOWED")
        base_url: str = os.path.join(
            os.environ["OIP_API_HOST"], cls.__server_module_name, dataset_entity_name
        )
        base_url = base_url.replace("\\", "/")
        headers: Dict = {"authorization": "APIKey " + os.environ["OIP_API_KEY"]}
        if method == "create":
            url: str = os.path.join(base_url, "add")
            response = HttpClient.post(url=url, json=data, headers=headers)
        elif method == "update":
            url = os.path.join(base_url, "update")
            response = HttpClient.put(url=url, json=data, headers=headers)

        if response.status_code == 200:
            return response.json()

    @classmethod
    @typechecked
    def get_data(
        cls,
        entity: str,
        api_request_args: ApiRequestArgs,
    ) -> Optional[EntityListInstances]:
        """
        Retrieve data from the API based on the provided parameters.

        :param entity: str: The name of the entity for which to retrieve the data.
        :param api_request_args:api_request_args: Additional parameters for the API request.
        :return: list: The retrieved data.
        """
        url: str = os.path.join(
            os.environ["OIP_API_HOST"], cls.__server_module_name, entity
        )
        url = url.replace("\\", "/")
        headers: Dict[str, str] = {
            "authorization": "APIKey " + os.environ["OIP_API_KEY"]
        }
        resp = HttpClient.get(url, headers, api_request_args)
        if resp.status_code == 200:
            return resp.json()["data"]
        return None

    @classmethod
    @typechecked
    def create_presigned_post_url(
        cls,
        object_name: str,
    ) -> Optional[Dict[str, Union[str, Dict[str, Union[str, Dict[str, str]]]]]]:
        """
        Generate a pre-signed URL for uploading an object to the cloud storage.
        :param object_name: str: The name of the object to create a pre-signed URL for.
        :return:dict: information about the presigned url
        """
        url: str = os.path.join(os.environ["OIP_API_HOST"], cls.__server_module_name)
        url = os.path.join(url, "presigned_post_url")
        url = url.replace("\\", "/")
        headers: Dict = {"authorization": "APIKey " + os.environ["OIP_API_KEY"]}
        data: Dict = {"object_name": object_name}
        resp = HttpClient.post(url=url, json=data, headers=headers)
        if resp.status_code == 200:
            return resp.json()
        return None

    @classmethod
    @typechecked
    def create_presigned_get_url(
        cls,
        object_name: str,
    ) -> Optional[str]:
        """
        Generate a pre-signed URL for downloading an object from the cloud storage.
        :param object_name: str: The name of the object to create a pre-signed URL for.
        :return:str: presigned get url
        """
        url: str = os.path.join(os.environ["OIP_API_HOST"], cls.__server_module_name)

        url = os.path.join(url, "presigned_get_url")
        url = url.replace("\\", "/")
        headers: Dict = {"authorization": "APIKey " + os.environ["OIP_API_KEY"]}
        data: Dict = {"object_name": object_name}
        resp = HttpClient.post(url=url, json=data, headers=headers)
        if resp.status_code == 200:
            return resp.json()
        return None

    @classmethod
    @typechecked
    def tag_dataset_as_main(
        cls,
        dataset_id: str,
        dataset_entity_name: str,
    ) -> None:
        """
        Tag current dataset as the main version, at the same time
        removing the tag from the previous main version.
        :param dataset_id: str: id of the dataset we are looking to tag as main
        :param dataset_entity_name: str: dataset entity name
        :return: none
        """
        url: str = os.path.join(
            os.environ["OIP_API_HOST"],
            cls.__server_module_name,
            dataset_entity_name,
            "tag_as_main",
            "id",
            dataset_id,
        )
        headers: Dict[str, str] = {
            "authorization": "APIKey " + os.environ["OIP_API_KEY"]
        }
        HttpClient.put(url=url, headers=headers)



import requests
import time
from typing import Dict, Optional, Union, Any
from .schema import DatasetDict
from .logger import Logger

logger = Logger().get_logger()


class HttpClient:
    MAX_RETRIES: int = 4

    @classmethod
    def get(
        cls,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        stream: Optional[bool] = False,
    ):
        """
        General get request function.
        Assigns headers and builds in retries and logging
        :param url: str: api host, example: https://www.example.com/api?x=1
        :param headers: Dict[str, str]: headers
        :param params: Dict[str, str]: params
        :param stream: bool:
        """
        """General 'make api request' function.
      Assigns headers and builds in retries and logging.
      """
        logger.debug(f"call http get method (url:{url})")
        base_log_record: Dict[str, Union[str, Dict[str, str]]] = dict(
            route=url, params=params
        )
        retry_count: int = 0

        while retry_count <= cls.MAX_RETRIES:
            start_time: float = time.time()
            try:
                response: Union[requests.Response, Exception] = requests.get(
                    url, params=params, headers=headers, timeout=None, stream=stream
                )
            except Exception as e:
                response = e

            elapsed_time: float = time.time() - start_time
            status_code: int = (
                response.status_code if hasattr(response, "status_code") else None
            )
            log_record: Dict[str, Union[int, float, str, Dict[str, str]]] = dict(
                base_log_record
            )
            log_record["elapsed_time_in_ms"] = 1000 * elapsed_time
            log_record["retry_count"] = retry_count
            log_record["status_code"] = status_code
            if status_code == 200:  # Success
                logger.debug("OK", extra=log_record)
                return response
            if status_code in [204, 206]:  # Success with a caveat - warning
                log_msg = {204: "No Content", 206: "Partial Content"}[status_code]
                logger.warning(log_msg, extra=log_record)
                return response
            log_record["tag"] = "failed_gro_api_request"
            if retry_count < cls.MAX_RETRIES:
                logger.warning(
                    response.text if hasattr(response, "text") else response,
                    extra=log_record,
                )
            if status_code in [400, 401, 402, 404, 301]:
                break  # Do not retry
            logger.warning("{}".format(response), extra=log_record)
            if retry_count > 0:
                # Retry immediately on first failure.
                # Exponential backoff before retrying repeatedly failing requests.
                time.sleep(2**retry_count)
            retry_count += 1
        raise APIError(response, retry_count, url)

    @classmethod
    def post(
        cls,
        url: str,
        data: Optional[Union[DatasetDict, Dict[str, str]]] = None,
        json=None,
        headers: Optional[Dict[str, str]] = None,
        files: Optional[Dict] = None,
        stream: Optional[bool] = False,
    ):
        """
        General post request function.
        Assigns headers and builds in retries and logging
        :param url: str: api host, example: https://www.example.com/api?x=1
        :param headers: Dict[str, str]: headers
        :param data: Dict[str, str]: params
        :param json: json data
        :param files: files
        :param stream: bool:
        """
        """General 'make api request' function.
     Assigns headers and builds in retries and logging.
     """
        logger.debug(f"call http post method (url {url})")
        base_log_record: Dict = dict(route=url, data=data)
        retry_count: int = 0

        while retry_count <= cls.MAX_RETRIES:
            start_time: float = time.time()
            try:
                if data:
                    response: Union[requests.Response, Exception] = requests.post(
                        url,
                        data=data,
                        headers=headers,
                        files=files,
                        timeout=None,
                        stream=stream,
                    )
                else:
                    response = requests.post(
                        url,
                        json=json,
                        headers=headers,
                        files=files,
                        timeout=None,
                        stream=stream,
                    )
            except Exception as e:
                response = e

            elapsed_time: float = time.time() - start_time
            status_code: int = (
                response.status_code if hasattr(response, "status_code") else None
            )
            log_record: Dict[str, Union[int, float, str, Dict[str, str]]] = dict(
                base_log_record
            )
            log_record["elapsed_time_in_ms"] = 1000 * elapsed_time
            log_record["retry_count"] = retry_count
            log_record["status_code"] = status_code
            if status_code in [200, 204]:  # Success
                logger.debug("OK", extra=log_record)
                return response
            log_record["tag"] = "failed_gro_api_request"
            if retry_count < cls.MAX_RETRIES:
                logger.warning(
                    response.text if hasattr(response, "text") else response,
                    extra=log_record,
                )
            if status_code in [400, 401, 402, 404, 301]:
                break  # Do not retry
            logger.warning("{}".format(response), extra=log_record)
            if retry_count > 0:
                # Retry immediately on first failure.
                # Exponential backoff before retrying repeatedly failing requests.
                time.sleep(2**retry_count)
            retry_count += 1
        raise APIError(response, retry_count, url)

    @classmethod
    def put(
        cls,
        url: str,
        json: DatasetDict = None,
        data: Any = None,
        headers: Optional[Dict[str, str]] = None,
        stream: Optional[bool] = False,
    ):
        """
        General put request function.
        Assigns headers and builds in retries and logging
        :param url: str: api host, example: https://www.example.com/api?x=1
        :param headers: Dict[str, str]: headers
        :param json: json data
        :param data: Dict[str, str]: params
        :param stream: bool:
        """
        """General 'make api request' function.
      Assigns headers and builds in retries and logging.
      """
        logger.debug(f"call http put method (url {url})")
        base_log_record: Dict = dict(route=url, data=json)
        retry_count: int = 0
        while retry_count <= cls.MAX_RETRIES:
            start_time: float = time.time()
            try:
                if json:
                    response: Union[requests.Response, Exception] = requests.put(
                        url, json=json, headers=headers, timeout=None, stream=stream
                    )
                else:
                    response = requests.put(
                        url, data=data, headers=headers, timeout=None, stream=stream
                    )
            except Exception as e:
                response = e

            elapsed_time: float = time.time() - start_time
            status_code: int = (
                response.status_code if hasattr(response, "status_code") else None
            )
            log_record: Dict[str, Union[int, float, str, Dict[str, str]]] = dict(
                base_log_record
            )
            log_record["elapsed_time_in_ms"] = 1000 * elapsed_time
            log_record["retry_count"] = retry_count
            log_record["status_code"] = status_code
            if status_code == 200:  # Success
                logger.debug("OK", extra=log_record)
                return response
            log_record["tag"] = "failed_gro_api_request"
            if retry_count < cls.MAX_RETRIES:
                logger.warning(
                    response.text if hasattr(response, "text") else response,
                    extra=log_record,
                )
            if status_code in [400, 401, 402, 404, 301]:
                break  # Do not retry
            logger.warning("{}".format(response), extra=log_record)
            if retry_count > 0:
                # Retry immediately on first failure.
                # Exponential backoff before retrying repeatedly failing requests.
                time.sleep(2**retry_count)
            retry_count += 1
        raise APIError(response, retry_count, url)


class APIError(Exception):
    def __init__(
        self,
        response,
        retry_count: int,
        url: str,
    ):
        self.response = response
        self.retry_count: int = retry_count
        self.url: str = url
        self.status_code = (
            response.status_code if hasattr(response, "status_code") else None
        )
        try:
            json_content = self.response.json()
            # 'error' should be something like 'Not Found' or 'Bad Request'
            self.message: str = json_content.get("error", "")
            # Some error responses give additional info.
            # For example, a 400 Bad Request might say "metricId is required"
            if "message" in json_content:
                self.message += ": {}".format(json_content["message"])
        except (Exception,):
            # If the error message can't be parsed, fall back to a generic "giving up" message.
            self.message = "Giving up on {} after {} {}: {}".format(
                self.url,
                self.retry_count,
                "retry" if self.retry_count == 1 else "retries",
                response,
            )



from typing import Optional
from ..HttpClient import HttpClient
from ..ServerFacade import ServerFacade
import tempfile
import json
from pathlib import Path
import os
import requests
from urllib.parse import urlparse
from typing import Dict, Union
import shutil
from time import time
from ..schema import DatasetState
from ..logger import Logger
from typeguard import typechecked

from .AwsProvider import AwsProvider
from .GcpProvider import GcpProvider
from .AzureProvider import AzureProvider
from .MinioProvider import MinioProvider

logger = Logger().get_logger()

COMPRESSED_EXTENSION = [
    ".zip",
    ".gz",
    ".tar.gz",
    ".bz2",
    ".7z",
    ".rar",
    ".tar",
    ".zst",
    ".xz",
    ".Z",
    ".sit",
    ".sitx",
]


class StorageManager:
    __default_target_folder = "global"
    # in MB
    __chunk_size = 5

    @staticmethod
    @typechecked
    def upload_state(object_name: str, dataset_state: DatasetState):
        """
        Upload state file to the cloud storage
        :param object_name: str: object name
        :param dataset_state: DatasetState: dataset state
        """
        logger.debug("call upload_state")
        presigned_post_data: Optional[
            Dict[str, Union[str, Dict[str, Union[str, Dict[str, str]]]]]
        ] = ServerFacade.create_presigned_post_url(object_name)
        if type(presigned_post_data) is dict:
            if type(presigned_post_data["kwargs"]) is dict:
                if type(presigned_post_data["kwargs"]["url"]) is str:
                    url: str = presigned_post_data["kwargs"]["url"]
                if (
                    "data" in presigned_post_data["kwargs"]
                    and type(presigned_post_data["kwargs"]["data"]) is dict
                ):
                    data: Dict[str, str] = presigned_post_data["kwargs"]["data"]
                else:
                    data = None
            if type(presigned_post_data["method"]) is str:
                upload_method: str = presigned_post_data["method"]

            with tempfile.NamedTemporaryFile(mode="w+") as temp_state_file:
                json.dump(dataset_state, temp_state_file, indent=4)
                temp_state_file.seek(0)
                if upload_method == "PUT":
                    binary_data = temp_state_file.read()
                    HttpClient.put(url=url, data=binary_data)
                elif upload_method == "POST":
                    file_ = {"file": (object_name, temp_state_file)}
                    HttpClient.post(
                        url=url,
                        data=data,
                        files=file_,
                    )

    @staticmethod
    @typechecked
    def upload(object_name: str, path: str) -> bool:
        """
        Upload data file to the cloud storage
        :param object_name: str: object name
        :param path: str: path to the file
        :return: bool:
        """
        presigned_post_data: Optional[
            Dict[str, Union[str, Dict[str, Union[str, Dict[str, str]]]]]
        ] = ServerFacade.create_presigned_post_url(object_name)
        if type(presigned_post_data) is dict:
            if type(presigned_post_data["kwargs"]) is dict:
                if type(presigned_post_data["kwargs"]["url"]) is str:
                    url: str = presigned_post_data["kwargs"]["url"]
                if (
                    "data" in presigned_post_data["kwargs"]
                    and type(presigned_post_data["kwargs"]["data"]) is dict
                ):
                    data: Dict[str, str] = presigned_post_data["kwargs"]["data"]
                else:
                    data = None
            if type(presigned_post_data["method"]) is str:
                upload_method: str = presigned_post_data["method"]
            with open(path, "rb") as file:
                if upload_method == "PUT":
                    binary_data = file.read()
                    HttpClient.put(url=url, data=binary_data)
                else:
                    file_ = {"file": (object_name, file)}

                    HttpClient.post(
                        url=url,
                        data=data,
                        files=file_,
                    )

            return True
        return False

    @staticmethod
    @typechecked
    def download(object_name: str, download_path: str) -> Optional[str]:
        """
        Download data file from the cloud storage
        :param object_name: str: object name
        :param download_path: str: download location
        :return: bool:
        """
        presigned_get_data: Optional[str] = ServerFacade.create_presigned_get_url(
            object_name
        )
        if type(presigned_get_data) is str:
            get_url: str = presigned_get_data
            response = HttpClient.get(get_url)
            if response.status_code == 200:
                with open(download_path, "wb") as file:
                    file.write(response.content)
                return download_path
        return None

    @staticmethod
    @typechecked
    def download_state(object_name: str) -> Optional[DatasetState]:
        """
        Download state file from the cloud storage
        :param object_name: str: object name
        :return: DatasetState: dataset state
        """
        presigned_get_data: Optional[str] = ServerFacade.create_presigned_get_url(
            object_name
        )
        if type(presigned_get_data) is str:
            get_url: str = presigned_get_data
            response = HttpClient.get(get_url)
            if response.status_code == 200:
                return response.json()
        return None

    @classmethod
    @typechecked
    def get_local_copy(
        cls,
        remote_url: str,
        target_folder: Optional[str] = None,
        extract_archive: Optional[bool] = True,
    ) -> str:
        """
        Get a local copy of a external dataset (dataset that doesn't belong to our storage)
        :param remote_url: str: dataset url
        :type target_folder: str: the local directory where the dataset will be downloaded.
        :param extract_archive: bool: if true extract the compressed file (defaults to True)
        :return: str: path to the downloaded file
        """
        # Parse the URL to extract the path, which includes the filename.
        url_parts = urlparse(remote_url)
        remote_path = url_parts.path
        file_name = remote_path.split("/")[-1]
        if not target_folder:
            target_folder = cls.__default_target_folder
        # we download into temp_local_path so that if we accidentally stop in the middle,
        # we won't think we have the entire file
        timestamp = str(time()).replace(".", "")
        temp_target_folder: str = f"{target_folder}_{timestamp}_partially"
        temp_dst_file_path: str = os.path.join(temp_target_folder, file_name)
        Path(temp_target_folder).mkdir(parents=True, exist_ok=True)

        with requests.get(url=remote_url, stream=True) as response:
            try:
                with open(temp_dst_file_path, mode="wb") as file:
                    for chunk in response.iter_content(
                        chunk_size=cls.__chunk_size * 1024
                    ):
                        file.write(chunk)
            except (Exception,):
                shutil.rmtree(temp_target_folder)
                raise Exception("failed in downloading file")

        suffix: str = Path(temp_dst_file_path).suffix.lower()
        if suffix == ".gz":
            suffix = "".join(a.lower() for a in Path(temp_dst_file_path).suffixes[-2:])
        if extract_archive and suffix in COMPRESSED_EXTENSION:
            try:
                unzipped_folder = file_name[: -len(suffix)]
                path_to_unzipped_temp_target_folder: str = (
                    f"{temp_target_folder}/{unzipped_folder}"
                )
                shutil.unpack_archive(
                    temp_dst_file_path, path_to_unzipped_temp_target_folder
                )
                os.remove(temp_dst_file_path)
                path_to_unzipped_target_folder = f"{target_folder}/{unzipped_folder}"
                count_suffix: int = 0
                # in case we already download the dataset
                while os.path.exists(path_to_unzipped_target_folder):
                    count_suffix += 1
                    path_to_unzipped_target_folder = (
                        f"{target_folder}/{unzipped_folder}-{str(count_suffix)}"
                    )
                shutil.move(path_to_unzipped_temp_target_folder, target_folder)
                shutil.rmtree(temp_target_folder)
                return path_to_unzipped_target_folder
            except Exception as e:
                shutil.rmtree(temp_target_folder)
                raise Exception(e)
        # not a compressed file
        else:
            dst_file_path: str = f"{target_folder}/{file_name}"
            if os.path.exists(dst_file_path):
                base_file_name: str = file_name[: -len(suffix)]
                dst_file_path = f"{target_folder}/{base_file_name}-1{suffix}"
                count_suffix = 1
                while os.path.exists(dst_file_path):
                    count_suffix += 1
                    dst_file_path = (
                        f"{target_folder}/{base_file_name}-{str(count_suffix)}{suffix}"
                    )
                os.rename(
                    temp_dst_file_path,
                    f"{temp_target_folder}/{base_file_name}-{count_suffix}{suffix}",
                )
                temp_dst_file_path = (
                    f"{temp_target_folder}/{base_file_name}-{count_suffix}{suffix}"
                )

                shutil.move(temp_dst_file_path, target_folder)

            else:
                os.rename(temp_dst_file_path, dst_file_path)
            shutil.rmtree(temp_target_folder)
            return dst_file_path

    @staticmethod
    @typechecked
    def add_aws_storage_conf(
        storage_conf_name: str, access_key: str, secret_key: str, region: str
    ):
        """
        Add AWS storage configuration. You can include one or multiple AWS configurations
            by recalling this function.
        :param storage_conf_name: str: storage configuration name serve as an identifier.
        :param access_key: str: aws access key
        :param secret_key: str: aws secret key
        :param region: str: aws region
        :return: none:
        """
        aws_conf: Dict[str, str] = {
            "PROVIDER": "AWS",
            "ACCESS_KEY": access_key,
            "SECRET_KEY": secret_key,
            "REGION": region,
        }

        os.environ[storage_conf_name] = json.dumps(aws_conf)

    @classmethod
    @typechecked
    def download_via_storage_provider(
        cls,
        storage_conf_name: str,
        download_uri: str,
        download_folder: str,
        max_workers: int,
        verbose: bool,
    ):
        """
        Download all the dataset files located in a specific download_uri
        :param storage_conf_name: str: storage configuration name serve as an identifier.
        :param download_uri: str: uri from where we download the dataset files refers to the specific location or path
            within the storage provider where the dataset files are hosted.
            The following are examples of ``download_uri``
             - S3: s3://bucket/folder/../dataset_folder
             - Google Cloud Storage: gs://bucket/folder/../dataset_folder
             - Azure Storage: azure://bucket/folder/../dataset_folder
             - Minio Storage: minio://bucket/folder/../dataset_folder
             - Other https://path_to_the_dataset_folder
        :param download_folder: str: destination directory  or folder where downloaded files will be stored.
        :param max_workers: int:Numbers of threads to be spawned when zipping and uploading the files.
            If None number of logical cores
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: str: path to the downloaded dataset directory
        """
        provider_mapping = {
            "AWS": AwsProvider,
            "GCP": GcpProvider,
            "AZURE": AzureProvider,
            "MINIO": MinioProvider,
        }
        # load storage conf
        storage_conf: Dict[str, str] = json.loads(os.environ[storage_conf_name])
        provider: str = storage_conf["PROVIDER"]
        if provider in provider_mapping:
            return provider_mapping[provider].download(
                storage_conf_name, download_uri, download_folder, max_workers, verbose
            )
        elif provider == "OTHER" and download_uri.startswith("http"):
            return cls.get_local_copy(download_uri, download_folder)
        else:
            raise ValueError(
                f"{provider} is not a valid provider please check the  official documentation"
            )



import logging
from typing import Any

LOGGING_FORMAT = "%(asctime)s - %(levelname)s - %(filename)s:%(lineno)d - %(message)s"
LOGGING_LEVEL = "INFO"


class Singleton(type):
    """
    Singleton metaclass
    """

    _instances: Any = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Logger(object, metaclass=Singleton):
    # __metaclass__ = SingletonType   # python 2 Style
    """
    Logger is the parent class of all the logger
    """
    _logger = None

    def __init__(self):
        self._logger = logging.getLogger("_logger")
        self._logger.setLevel(LOGGING_LEVEL)
        formatter = logging.Formatter(LOGGING_FORMAT)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        self._logger.addHandler(stream_handler)

    def get_logger(self):
        return self._logger



import os
from ..dataset import DatasetClient

api_host = "http://192.168.1.35:8002"
workspace_name = "default_workspace"
api_key = "4fc086fe-d64b-4ac8-816a-0dcbf34f7368"
workspace_id = "8fb7c66a-a49d-46a8-8044-51a94cee1a36"
DatasetClient.connect(api_host=api_host, api_key=api_key, workspace_name=workspace_name)
assert os.environ["OIP_API_HOST"] == api_host
assert os.environ["WORKSPACE_ID"] == workspace_id
assert os.environ["ML_DATASET_AUTHENTICATION_TOKEN"] == api_key
# we need to create it first
dataset = DatasetClient.create(name="my_dataset")
dataset.add_files(
    "/Users/mohamed/projects/appbuilder/client/ml_dataset_client/examples/countries_dataset/europe"
)
dataset.upload()
dataset.finalize()


dataset = DatasetClient.get(dataset_name="north_america_dataset")
dataset.get_local_copy()



from setuptools import setup, find_packages

setup(
    name="oip-dataset-client",
    version="0.0.2",
    author="Rachid Belmeskine",
    author_email="rachid.belmeskine@gmail.com",
    description="This is the API client of Open Innovation MLOps Platform - Dataset",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    # url='https://github.com/...',
    python_requires=">=3.7",
    packages=find_packages(),
    install_requires=[
        "typeguard==2.13.3",
        "requests==2.27.1",
        "six==1.16.0",
        "psutil==5.9.8",
        "tqdm==4.66.1",
        "boto3==1.26.20",
        "PyYAML==6.0.1",
        "attrs==23.2.0",
        "python-dotenv==1.0.1"
    ],
    license="PRIVATE LICENSE",
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "console_scripts": [
            "oip-dataset-client-init = oip_dataset_client.cli.config.__main__:main",
        ],
    },
)



import os
from .version import Version
from .Storage.StorageManager import StorageManager
from .schema import (
    FileEntryDict,
    ApiRequestArgs,
    EntityListInstances,
    DatasetDict,
    DatasetState,
    FileEntries,
)
from copy import deepcopy
from enum import Enum
from tqdm import tqdm
from typing import Optional, List, Dict, Union, Set, Tuple
from typeguard import typechecked
from pathlib import Path
from .utils import (
    check_connection,
    check_storage_conf,
    sha256sum,
    matches_any_wildcard,
)
from multiprocessing.pool import ThreadPool
import psutil
import uuid
from copy import copy
from .ServerFacade import ServerFacade
from concurrent.futures import ThreadPoolExecutor
from .ParallelZipper import ParallelZipper
from zipfile import ZIP_DEFLATED
import shutil
from time import time
from .logger import Logger
from functools import partial
from .ServerConf import DATASET_ENTITY_NAME, WORKSPACE_ENTITY_NAME
import tempfile

logger = Logger().get_logger()


class FileEntry(object):
    """
    This class aids in the retrieval and upload of files associated with a dataset.
    It allows us to determine if a file has been successfully uploaded to cloud the storage, identifies
    the parent dataset to which the file belongs, and provides essential details for downloading the file.
    """

    def __init__(
        self,
        relative_path: str,
        parent_dataset_id: str,
        artifact_name: Optional[str] = None,
        local_path: Optional[str] = None,
        size: Optional[int] = None,
        hash_: Optional[str] = None,
        remote_url_basename: Optional[str] = None,
    ) -> None:
        """
        Initialize a FileEntry object.

        :param relative_path: str: The relative path of the file within the dataset.
        :param parent_dataset_id: str: The unique identifier of the parent dataset.
            the dataset that the file belongs
        :param artifact_name: str: The name of the artifact that the file belongs.
            All files with the same parent dataset will be stored in chunks of the same size
            and each chunk corresponds to an artifact. The naming of artifacts follows a sequential pattern
            with the first artifact named "data," and subsequent artifacts named "data_1," "data_2," and so on.
        :param local_path: str: The local file path.
        :param size: int: The size of the file in bytes.
        :param hash_: str: The hash value of the file.
            We use it to check whether the file has been modified or  not
        :param remote_url_basename: str: The basename of the remote URL.
            In general, when we upload files to cloud storage the remote base URL serves
            as the file's name on the cloud storage platform.
        """
        self.artifact_name = artifact_name
        self.relative_path = relative_path
        self.parent_dataset_id = parent_dataset_id
        self.size = size
        self.hash = hash_
        # cleared when file is uploaded.
        self.local_path = local_path
        # added when we upload the file
        self.remote_url_basename = remote_url_basename

    def serializer(self) -> FileEntryDict:
        """
        Serializes the object data into dict
        :return: dict:
        """
        state: FileEntryDict = {
            "artifact_name": self.artifact_name,
            "relative_path": self.relative_path,
            "hash": self.hash,
            "parent_dataset_id": self.parent_dataset_id,
            "size": self.size,
        }
        state["remote_url_basename"] = (
            self.remote_url_basename if self.remote_url_basename else None
        )
        state["local_path"] = self.local_path if self.local_path else None
        return state


class DatasetStatusEnum(Enum):
    def __str__(self) -> str:
        return str(self.value)

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    created: str = "created"
    in_progress: str = "in_progress"
    uploaded: str = "uploaded"
    failed: str = "failed"
    completed: str = "completed"
    aborted: str = "aborted"


STATUS_FLOW = {
    DatasetStatusEnum.created.value: [
        DatasetStatusEnum.in_progress.value,
        DatasetStatusEnum.completed.value,
        DatasetStatusEnum.aborted.value,
        DatasetStatusEnum.failed.value,
    ],
    DatasetStatusEnum.in_progress.value: [
        DatasetStatusEnum.uploaded.value,
        DatasetStatusEnum.aborted.value,
        DatasetStatusEnum.failed.value,
    ],
    DatasetStatusEnum.uploaded.value: [
        DatasetStatusEnum.completed.value,
        DatasetStatusEnum.in_progress.value,
        DatasetStatusEnum.aborted.value,
        DatasetStatusEnum.failed.value,
    ],
}


class DatasetClient(object):
    """
    The core class of the ML dataset client that provides a comprehensive set of methods
    to effectively manage datasets. These methods include functions for adding and removing
    files, uploading and downloading data, and finalizing dataset operations.
    This class serves as the central interface for interacting with machine learning datasets
    """

    __dataset_entity_name: str = DATASET_ENTITY_NAME
    __workspace_entity_name: str = WORKSPACE_ENTITY_NAME
    __default_dataset_version: str = "1.0.0"
    # The cloud storage directory name where the datasets will be stored.
    __datasets_entry_name: str = "datasets"
    # The name of teh cloud storage directory that will hold the state file.
    # A state file contains information about all the file entries within the dataset.
    # example state-> state.json
    __state_entry_name: str = "state"
    # The cloud storage parent directory of __datasets_entry_name example: data->datasets
    __default_data_entry_name: str = "data"
    # we us it as a prefix when naming artifacts example data_1,data_2....
    __data_entry_name_prefix: str = "data_"
    # Artifact chunk size (MB) for the compressed dataset.
    _dataset_chunk_size_mb: int = 512
    __dataset_platform: str = "oip"
    # The local directory name where the datasets will be stored.
    __dataset_download_folder_context: str = "datasets"
    # The directory prefix for storing downloaded dataset files.
    # It always follows the format 'prefix_{UUID of the dataset}'.
    __dataset_download_folder_prefix: str = "ds"
    verbose: bool = True

    @typechecked
    def __init__(
        self,
        name: str,
        id_: str = None,
        workspace_id: Optional[str] = None,
        version: Optional[str] = None,
        is_main: bool = False,
        parent_datasets: List[str] = None,
        status: Optional[str] = None,
        file_count: int = 0,
        added_files: int = 0,
        modified_files: int = 0,
        removed_files: int = 0,
        total_size: int = 0,
        total_size_compressed: Optional[int] = 0,
        description: str = None,
        tags: List[str] = None,
    ):
        """
        Create a new dataset, Do not use directly! Use Dataset.create(...).
        :param name: str: Naming the new dataset.
        :param id_: str: Id of the new Dataset.
        :param workspace_id: str: The workspace ID to which the new dataset will belong.
        :param version: str: Version of the new dataset. If not set, try to find the latest version
            of the dataset with given `name`  and auto-increment it.
        :param is_main: bool: True if this version is the main version
        :param parent_datasets: list[str]: List of parent datasets (parent ids)
        :param status: str: Status of the new-dataset
            status in [created,in_progress,uploaded,completed,aborted,failed,unknown]
        :param file_count: int: The total count of files within the dataset.
        :param added_files: int: The total count of files added to the dataset
            Added files refer to those that are not part of the direct or indirect
            parent datasets of the new dataset.
        :param modified_files: int: The total count of files modified within the dataset.
            Modified files are files that belong to the direct or indirect parent datasets
            and we upload a modified version of these files to the new dataset.
        :param removed_files: int: The total count of files removed from the new dataset.
            Removed files are files that belong to the direct or indirect parent datasets
            ,and we have removed them from the new dataset.
        :param total_size: int: The total size of all the files within the new dataset
            including both the direct and indirect parent files.
        :param total_size_compressed: int: The total compressed size of all the files within the new dataset
            including both the direct and indirect parent files.
        :param description: str: Description of the dataset
        :param tags: list[str]: Descriptive tags categorize datasets by keywords, detailing their subject
            matter, domain, or specific topics for better identification and organization.
        :return: Newly created Dataset object
        """

        self._name: str = name
        self._id: str = id_ if id_ else str(uuid.uuid4())
        self._status: str = DatasetStatusEnum.created.value if not status else status
        self._version: str = version
        self._is_main: bool = is_main
        if not parent_datasets:
            parent_datasets = []
        self._parent_datasets: List[str] = parent_datasets
        self._workspace_id: str = workspace_id or os.environ["OIP_WORKSPACE_ID"]

        self._description: str = description
        if not tags:
            tags = []
        self._tags: List[str] = tags
        self._file_count: int = file_count
        self._added_files: int = added_files
        self._modified_files: int = modified_files
        self._removed_files: int = removed_files
        self._total_size: int = total_size
        self._total_size_compressed: int = total_size_compressed

        self._file_entries: Dict[str, FileEntry] = {}
        self._dependency_graph: Dict[str, List[str]] = {}
        # set current artifact name to be used (support for multiple upload sessions)
        self._data_artifact_name: str = self._get_next_data_artifact_name()
        # store a cached lookup of the number of chunks each parent dataset has.
        self._dependency_chunk_lookup: Optional[Dict[str, int]] = None

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version

    @property
    def is_main(self) -> bool:
        return self._is_main

    @property
    def parent_datasets(self) -> List[str]:
        return self._parent_datasets

    @property
    def tags(self) -> List[str]:
        return self._tags

    @property
    def workspace_id(self) -> str:
        return self._workspace_id

    @version.setter
    @typechecked
    def version(self, version: str) -> None:
        """
        Version Setter
        :param version: str:
        :return: str:
        """
        version = str(version).strip()

        if not Version.is_valid_version_string(version):
            raise ValueError(f"{version} is not a valid version")
        else:
            self._version = version

    @property
    def total_size(self) -> int:
        return self._total_size

    @property
    def total_size_compressed(self) -> int:
        return self._total_size_compressed

    @property
    def file_entries(self) -> List[FileEntry]:
        return list(self._file_entries.values())

    @property
    def file_count(self) -> int:
        return self._file_count

    @property
    def added_files(self) -> int:
        return self._added_files

    @property
    def modified_files(self) -> int:
        return self._modified_files

    @property
    def removed_files(self) -> int:
        return self._removed_files

    @property
    def file_entries_dict(self) -> Dict[str, FileEntry]:
        """
        File entries ad dict
        :return: dict with relative file path as key, and FileEntry as value
        """
        return self._file_entries

    @typechecked
    def _update_status(self, new_status: str) -> None:
        """
        Update Dataset status
        :param new_status: str: new status
        :return: none
        """
        if new_status not in STATUS_FLOW[self._status]:
            self._save(
                "update", {"id": self._id, "status": DatasetStatusEnum.failed.value}
            )
            raise ValueError(f"forbidden flow  form {self._status} to {new_status}")
        self._status = new_status

    @typechecked
    def _get_next_data_artifact_name(
        self, last_data_artifact_name: Optional[str] = None
    ) -> str:
        """
        Get the next data artifact name
        :param last_data_artifact_name: str: name of the last data artifact example data_x
        :return: str:the next data artifact name example data_x+1
        """
        if not last_data_artifact_name:
            # the default value of the first data artifact
            return self.__default_data_entry_name
        else:
            prefix: str = self.__data_entry_name_prefix
            last_artificat_version: int = (
                int(last_data_artifact_name[len(prefix) :])
                if last_data_artifact_name.startswith(prefix)
                else 0
            )

        return f"{prefix}{last_artificat_version+1:03d}"

    @staticmethod
    @typechecked
    def connect(api_host: str, api_key: str, workspace_name: str) -> None:
        """
        Connect to the remote MLDataset platform
        It will check if the workspace exists and get its id
        Then set the the api host, the workspace id, the workspace_name, and the api key
        :param api_host: str: The API Host
        :param api_key: str: The API Key
        :param workspace_name: str: The workspace name
        :return: none:
        """

        # Check if workspace exists
        api_request_args: ApiRequestArgs = {
            "filter_cols": "name",
            "filter_ops": "=",
            "filter_vals": workspace_name,
        }
        os.environ["OIP_API_HOST"] = api_host
        os.environ["OIP_API_KEY"] = api_key
        data = ServerFacade.get_data(
            entity=WORKSPACE_ENTITY_NAME,
            api_request_args=api_request_args,
        )
        if not data:
            raise ValueError(f"WORKSPACE {workspace_name} DOES NOT EXIST")
        workspace_id: str = data[0]["id"]
        os.environ["OIP_WORKSPACE_ID"] = workspace_id
        os.environ["OIP_WORKSPACE_NAME"] = workspace_name

    @classmethod
    @check_connection
    @typechecked
    def create(
        cls,
        name: str,
        parent_datasets: List[str] = None,
        version: str = None,
        is_main: bool = False,
        description: str = None,
        tags: List[str] = None,
        verbose: Optional[bool] = None,
    ):
        """
        Create a new dataset
        :param name: str: Naming the new dataset.
        :param version: str: version of the new dataset. If not set, try to find the latest version
        :param parent_datasets: list[str]: A list of parent datasets to extend the new dataset
            by adding all the files from their respective parent datasets
        :param is_main: boolean: True if this is the main dataset
        :param description: str: Description of the dataset
        :param parent_datasets: list[str]: Descriptive tags categorize datasets by keywords, detailing their
            subject matter, domain, or specific topics for better identification and organization.
        :param tags: List[str]: List of tags to be associated with the dataset
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: Newly created Dataset object
        """
        logger.debug("call dataset create")

        if verbose is None:
            verbose = cls.verbose

        if name == "":
            raise ValueError("`dataset name` can't be an empty string")
        if not parent_datasets:
            parent_datasets = list()

        parent_datasets_: List[DatasetClient] = [
            cls.get(dataset_id=p_id) for p_id in (parent_datasets or [])
        ]
        is_first_version: bool = cls._is_first_version(name)
        for p in parent_datasets_:
            if not is_first_version and p.name != name:
                raise ValueError("can't inherit from a parent with a different name")

            if not p.is_completed():
                raise ValueError(
                    "can't inherit from a parent that was not finalized/closed"
                )

        if version:
            version = str(version).strip()
            if not Version.is_valid_version_string(version):
                raise ValueError(f"{version} is not a valid version")
        if not tags:
            tags = []

        # make sure that the first version is always the main one
        if not is_main and is_first_version:
            is_main = True
        dataset: DatasetClient = cls(
            name=name,
            version=version,
            is_main=is_main,
            parent_datasets=parent_datasets,
            description=description,
            tags=tags,
        )

        if not dataset._version:
            highest_version: Optional[str] = dataset._get_highest_version()
            if highest_version is not None:
                dataset._version = str(Version(highest_version).get_next_version())
            else:
                dataset._version = cls.__default_dataset_version
        # merge datasets according to order
        dependency_graph: Dict[str, List[str]] = {}
        parent_file_entries: Dict[str, FileEntry] = {}
        total_size: int = 0
        for p in parent_datasets_:
            parent_file_entries.update(deepcopy(p._file_entries))
            dependency_graph.update(deepcopy(p._dependency_graph))
            total_size += p.total_size

        # key for the dataset file entries are the relative path within the data
        dataset._file_entries = parent_file_entries
        dataset._file_count = len(dataset.file_entries)
        dataset._total_size = total_size
        # this will create a graph of all the dependencies we have, each entry lists it's own direct parents
        dataset._dependency_graph = dependency_graph
        dataset._dependency_graph[dataset._id] = [p._id for p in parent_datasets_]
        dataset._update_dependency_graph()
        data: DatasetDict = dataset.serializer()
        # Fields that are excluded from being stored on the server.
        excluded_fields: List[str] = [
            "dependency_graph",
            "file_entries",
            "dependency_chunk_lookup",
        ]
        for field in excluded_fields:
            data.pop(field)
        # Upload the state file to the cloud storage.
        dataset._upload_state()
        # Store the data on the sever
        cls._save(method="create", data=data)
        if verbose:
            logger.info("THE DATASET HAS BEEN SUCCESSFULLY CREATED")
        return dataset

    @check_connection
    def tag_as_main(self) -> None:
        """
        Tag current dataset as the main version, at the same time
        removing the tag from the previous main version.
        :return: none
        """
        ServerFacade.tag_dataset_as_main(
            self._id,
            self.__dataset_entity_name,
        )

    def _get_highest_version(self) -> Optional[str]:
        """
        Get the highest semantic version of the dataset
        :return: union[str,none]: the highest version of the dataset, or none if it doesn't exist.
        """
        workspace_id: str = os.environ["OIP_WORKSPACE_ID"]

        api_request_args: ApiRequestArgs = {
            "filter_cols": f"name|{self.__workspace_entity_name}_id",
            "filter_ops": "=|=",
            "filter_vals": f"{self.name}|{workspace_id}",
        }
        datasets: EntityListInstances = ServerFacade.get_data(
            self.__dataset_entity_name,
            api_request_args,
        )
        versions: List[str] = [d["version"] for d in datasets]
        versions = sorted(
            versions,
            key=lambda v: [
                int(part) if part.isdigit() else part for part in v.split(".")
            ],
        )
        return versions[-1] if versions else None

    @classmethod
    @typechecked
    def _is_first_version(cls, dataset_name: str) -> bool:
        """
        Check if the dataset with the specified name is the first version of the dataset.
        :param dataset_name: str:  dataset name we are looking for
        :return: bool: true if the dataset with the specified name is the first version of the dataset
        """
        workspace_id: str = os.environ["OIP_WORKSPACE_ID"]

        api_request_args: ApiRequestArgs = {
            "filter_cols": f"name|{cls.__workspace_entity_name}_id",
            "filter_ops": "=|=",
            "filter_vals": f"{dataset_name}|{workspace_id}",
        }
        data: EntityListInstances = ServerFacade.get_data(
            cls.__dataset_entity_name,
            api_request_args,
        )
        return len(data) == 0

    @classmethod
    @check_connection
    @typechecked
    def get(
        cls,
        dataset_id: Optional[str] = None,
        dataset_name: Optional[str] = None,
        dataset_version: Optional[str] = None,
        only_completed: Optional[bool] = False,
        auto_create: Optional[bool] = False,
    ):
        """
        Get a specific Dataset. If multiple datasets are found, the dataset with the
        highest semantic version is returned. This functions raises an Exception in case no dataset
        can be found and the ``auto_create=True`` flag is not set

        :param dataset_id: str: Requested dataset ID
        :param dataset_name: str: Requested dataset name
        :param dataset_version: str: Requested version of the Dataset
        :param only_completed: str: Return only if the requested dataset is completed
        :param auto_create: bool: Create a new dataset if it does not exist yet
        :return: Dataset object
        """

        if not any([dataset_id, dataset_name]):
            raise ValueError(
                "Dataset selection criteria not met. Didn't provide id/name correctly."
            )

        if dataset_id:
            api_request_args: ApiRequestArgs = {
                "filter_cols": "id",
                "filter_ops": "=",
                "filter_vals": f"{dataset_id}",
                "sort_col": "version",
                "sort_order": "-1",
            }
        else:
            workspace_id: str = os.environ["OIP_WORKSPACE_ID"]
            if not dataset_version:
                api_request_args = {
                    "filter_cols": "name|workspace_id",
                    "filter_ops": "=|=",
                    "filter_vals": f"{dataset_name}|{workspace_id}",
                    "sort_col": "version",
                    "sort_order": "-1",
                }

            else:
                api_request_args = {
                    "filter_cols": "name|workspace_id|version",
                    "filter_ops": "=|=|=",
                    "filter_vals": f"{dataset_name}|{workspace_id}|{dataset_version}",
                }
        if only_completed:
            api_request_args["filter_cols"] += "|status"
            api_request_args["filter_ops"] += "|="
            api_request_args["filter_vals"] += f"|{DatasetStatusEnum.completed.value}"

        data: Optional[EntityListInstances] = ServerFacade.get_data(
            cls.__dataset_entity_name,
            api_request_args,
        )
        if data:
            # return the latest version
            dataset: DatasetClient = cls._deserializer([data[0]])[0]
            return dataset
        elif not data and auto_create and dataset_name:
            dataset = cls.create(name=dataset_name, version=dataset_version)
            return dataset
        else:
            raise ValueError(
                f"Could not find Dataset {'id' if dataset_id else 'name/version'} "
                f"{dataset_id if dataset_id else (dataset_name, dataset_version)}"
            )

    @typechecked
    def get_local_copy(
        self,
        parent_folder: Optional[str] = None,
        max_workers: Optional[int] = None,
        verbose: Optional[bool] = None,
    ) -> str:
        """
        Get a local copy of the entire dataset
        :param max_workers: int: Number of threads to be spawned when getting the dataset copy.
            Defaults to the number of logical cores.
        :param parent_folder: str: path to the parent folder
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: str: path to the downloaded folder
        """
        if verbose is None:
            verbose = self.verbose

        if not self.is_completed():
            raise ValueError(
                "can't get a local copy of a dataset that was not finalized"
            )
        max_workers = max_workers or psutil.cpu_count()
        target_base_folder: str = self._create_ds_target_base_folder(parent_folder)
        # we download into temp_local_path so if we accidentally stop in the middle,
        # we won't think we have the entire dataset
        timestamp = str(time()).replace(".", "")
        temp_target_base_folder = f"{target_base_folder}_{timestamp}_partially"
        os.rename(target_base_folder, temp_target_base_folder)
        artifacts: Dict[str, List[Dict]] = self._get_artifacts_to_download()
        download_prog_bar: tqdm = None
        if verbose:
            total_artifacts = sum(len(l_art) for l_art in artifacts.values())
            download_prog_bar = tqdm(
                total=total_artifacts, desc="downloading chunks", unit=" chunk"
            )
        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            for dataset_id in artifacts:
                if self._id == dataset_id:
                    dataset_to_download = self
                else:
                    dataset_to_download = self.get(dataset_id)
                self._get_dataset_files_from_artifacts(
                    dataset_to_download.name,
                    dataset_to_download.id,
                    artifacts[dataset_to_download.id],
                    temp_target_base_folder,
                    pool,
                    download_prog_bar,
                )

        os.rename(temp_target_base_folder, target_base_folder)
        return target_base_folder

    @classmethod
    @typechecked
    def _get_dataset_files_from_artifacts(
        cls,
        dataset_name: str,
        dataset_id: str,
        artifacts: List[Dict],
        target_folder: str,
        pool: ThreadPoolExecutor = None,
        download_prog_bar: tqdm = None,
    ) -> str:
        """
        Retrieve all dataset files from a specific artifacts, which means obtaining all
        the files associated with a particular chunks.
        :param dataset_id: str: dataset id
        :param dataset_name: str: dataset name
        :param artifacts: list[Dict]: list of artifacts names
            example :[{artifact_name:"..",remote_url_basename:".."},..{..}]
        :param target_folder: str: path of the  Target folder for the download files
        :param pool: ThreadPoolExecutor: A thread pool for concurrent execution of tasks.
        :return: str: path to the target folder
        """
        for art in artifacts:
            remote_url: str = cls.generate_object_name(
                dataset_name=dataset_name,
                dataset_id=dataset_id,
                artifact_name=art["artifact_name"],
                file_path=art["remote_url_basename"],
            )

            download_path: str = os.path.join(target_folder, art["remote_url_basename"])
            download_task = pool.submit(
                StorageManager.download, remote_url, download_path
            )

            download_task.add_done_callback(
                partial(
                    DatasetClient.unzip_file_after_download,
                    download_prog_bar=download_prog_bar,
                    target_folder=target_folder,
                )
            )

        Path(target_folder).touch()

        return target_folder

    @staticmethod
    @typechecked
    def unzip_file_after_download(
        task, target_folder: str, download_prog_bar: tqdm = None
    ) -> None:
        download_path: str = task.result()
        shutil.unpack_archive(download_path, target_folder)
        os.remove(download_path)
        if download_prog_bar:
            download_prog_bar.update(1)

    def _create_ds_target_base_folder(self, parent_folder: Optional[str] = None) -> str:
        """
        Create a target base folder for storing the downloaded dataset files
        :param parent_folder: str: path to the parent folder
        :return: str: path of the target folder
        """

        target_base_folder: str = os.path.join(
            self.__dataset_platform,
            self.__dataset_download_folder_context,
            self._name,
            f"{self.__dataset_download_folder_prefix}_{self._id}",
        )
        if parent_folder:
            target_base_folder = os.path.join(parent_folder, target_base_folder)
        target_folder: str = target_base_folder
        suffix: int = 1
        # check if the folder already exist if yes we add a suffix
        while os.path.exists(target_folder):
            target_folder = f"{target_base_folder}-{str(suffix)}"
            suffix += 1
        Path(target_folder).mkdir(parents=True)
        return target_folder

    def _get_parents(self) -> List[str]:
        """
        Return a list of direct parent datasets (str)
        :return: list[str]: list of dataset ids
        """
        return self._dependency_graph[self.id]

    @classmethod
    @typechecked
    def _deserializer(cls, data: EntityListInstances):
        """
        Deserialize a list of dictionaries into a list of dataset objects.
        :param data:EntityListInstances:
        :return: list: return a list dataset objects
        """
        datasets: List[DatasetClient] = []
        for d in data:
            for key in copy(d):
                if key.startswith("_"):
                    del d[key]
                elif key in [
                    "file_count",
                    "added_files",
                    "removed_files",
                    "modified_files",
                    "total_size",
                    "total_size_compressed",
                ]:
                    # cast float to int
                    # when we get the data from the server we get is as float
                    d[key] = int(d[key])
            # Rename 'id' to 'id_'
            d["id_"] = d.pop("id")
            dataset: DatasetClient = cls(**d)
            dataset_state: DatasetState = dataset._download_state()
            dataset._dependency_graph = dataset_state["dependency_graph"]
            dataset._dependency_chunk_lookup = dataset_state["dependency_chunk_lookup"]
            for f in dataset_state["file_entries"]:
                # rename hash to hash_
                f["hash_"] = f.pop("hash")
                dataset._file_entries[f["relative_path"]] = FileEntry(**f)
            datasets.append(dataset)

        return datasets

    @check_connection
    @typechecked
    def add_files(
        self,
        path: str,
        wildcard: Optional[Union[str, List[str]]] = None,
        recursive: Optional[bool] = True,
        max_workers: Optional[int] = None,
        verbose: Optional[bool] = None,
    ) -> Tuple[int, int]:
        """
        Add files to the dataset.
         To add files to the dataset, we follow these steps:
           1-Calculate the file hash.
           2-Compare it against the parent dataset.
           3-If the file doesn't exist or has been modified, add it to the list of files to be uploaded later.
        :param path: str: path to the files we want to add to our dataset
        :param wildcard: str: add only specific set of files.
            Wildcard matching, can be a single string or a list of wildcards
        :param recursive: bool: If True, match all wildcard files recursively
        :param max_workers: int: The number of threads to add the files with. Defaults to the number of logical cores
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: tuple[int,int]: return (added_files,modified_files)
        """
        if self.is_final():
            raise Exception(
                "Files can't be added to a dataset that has already been finalized."
            )
        logger.debug("call add_files")

        if verbose is None:
            verbose = self.verbose

        if verbose:
            logger.info("ADDING FILES ...")
        max_workers = max_workers or psutil.cpu_count()
        added_files, modified_files = self._add_files(
            path=path,
            wildcard=wildcard,
            recursive=recursive,
            max_workers=max_workers,
            verbose=verbose,
        )

        self._update_file_entries_post_processing(
            removed_operation=False, verbose=verbose
        )
        if self._status == DatasetStatusEnum.created.value and (
            added_files or modified_files
        ):
            self._update_status(new_status=DatasetStatusEnum.in_progress.value)
        data_to_be_updated: DatasetDict = {
            "id": self._id,
            "added_files": self._added_files,
            "removed_files": self._removed_files,
            "modified_files": self._modified_files,
            "total_size": self.total_size,
            "total_size_compressed": self.total_size_compressed,
            "file_count": self.file_count,
            "status": self._status,
        }
        self._upload_state()
        self._save(
            method="update",
            data=data_to_be_updated,
        )
        if verbose:
            logger.info("FILES ADDED SUCCESSFULLY")
        return added_files, modified_files

    @typechecked
    def _add_files(
        self,
        path: str,
        verbose: bool,
        wildcard: Optional[Union[str, List[str]]] = None,
        recursive: Optional[bool] = True,
        max_workers: Optional[int] = None,
    ) -> Tuple[int, int]:
        """
        Add files to the dataset.
         To add files to the dataset, we follow these steps:
           1-Calculate the file hash.
           2-Compare it against the parent dataset.
           3-If the file doesn't exist or has been modified, add it to the list of files to be uploaded later.
        :param path: str: path to the files we want to add to our dataset
        :param verbose: bool: If True, enable console output for progress information
        :param wildcard: str: add only specific set of files.
            Wildcard matching, can be a single string or a list of wildcards
        :param recursive: bool: If True, match all wildcard files recursively
        :param max_workers: int: The number of threads to add the files with. Defaults to the number of logical cores
        :return: tuple[int,int]: return (added_files,modified_files)
        """
        max_workers = max_workers or psutil.cpu_count()
        path_: Path = Path(path)
        wildcard = wildcard or ["*"]
        if isinstance(wildcard, str):
            wildcard = [wildcard]
        # single file, no need for threading
        if path_.is_file():
            file_entry = self._calc_file_hash(
                FileEntry(
                    local_path=path_.absolute().as_posix(),
                    relative_path=path_.relative_to(path_.parent).as_posix(),
                    parent_dataset_id=self._id,
                )
            )
            file_entries: List = [file_entry]
        else:
            # if not a folder raise exception
            if not path_.is_dir():
                self._save(
                    "update", {"id": self._id, "status": DatasetStatusEnum.failed.value}
                )
                raise ValueError(f"Could not find file/folder { path_.as_posix()}")

            # prepare a list of files
            file_entries = []
            for w in wildcard:
                files: List[Path] = (
                    list(path_.rglob(w)) if recursive else list(path_.glob(w))
                )
                file_entries.extend([f for f in files if f.is_file()])
            file_entries = list(set(file_entries))
            file_entries = [
                FileEntry(
                    parent_dataset_id=self._id,
                    local_path=f.absolute().as_posix(),
                    relative_path=f.relative_to(path_).as_posix(),
                )
                for f in file_entries
            ]

            pool: ThreadPool = ThreadPool(max_workers)
            try:
                # Displaying progress bars when calculating file hashes
                for _ in tqdm(
                    pool.imap_unordered(self._calc_file_hash, file_entries),
                    desc="files processing",
                    unit=" file",
                    total=len(file_entries),
                    disable=not verbose,
                ):
                    pass
            except ImportError:
                pool.map(self._calc_file_hash, file_entries)
            pool.close()

        # merge back into the dataset
        added_files: int = 0
        modified_files: int = 0
        # Initialize progress bar for adding file entries
        for f in tqdm(
            file_entries, desc="adding files", unit=" file", disable=not verbose
        ):
            ds_cur_f: FileEntry = self._file_entries.get(f.relative_path)
            if not ds_cur_f:
                self._file_entries[f.relative_path] = f
                added_files += 1
            elif ds_cur_f.hash != f.hash:
                self._file_entries[f.relative_path] = f
                modified_files += 1
            elif (
                f.parent_dataset_id == self._id
                and ds_cur_f.parent_dataset_id == self._id
            ):
                # check if we have the file in an already uploaded chunk
                if ds_cur_f.local_path:
                    self._file_entries[f.relative_path] = f
                    added_files += 1
        return added_files, modified_files

    @check_connection
    @typechecked
    def remove_files(
        self,
        wildcard_path: str,
        recursive: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> int:
        """
        Remove files from the current dataset
        :param wildcard_path: str: wildcard path to the files to remove.
            The path is always relative to the dataset (e.g 'folder/file.bin').
        :param recursive: bool: If True, match all wildcard files recursively
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: int: Number of files removed
        """

        if self.is_final():
            raise Exception(
                "Files can't be removed from a dataset that has already been finalized."
            )
        if verbose is None:
            verbose = self.verbose

        if verbose:
            logger.info("REMOVING FILES ...")
        if wildcard_path and wildcard_path.startswith("/"):
            wildcard_path = wildcard_path[1:]
        self._file_entries = {
            relative_path: v
            for relative_path, v in self._file_entries.items()
            if not matches_any_wildcard(
                relative_path, wildcard_path, recursive=recursive
            )
        }
        self._update_file_entries_post_processing(verbose=verbose)
        if self._status == DatasetStatusEnum.created.value:
            self._update_status(new_status=DatasetStatusEnum.in_progress.value)
        data_to_be_updated: DatasetDict = {
            "id": self._id,
            "removed_files": self._removed_files,
            "added_files": self._added_files,
            "modified_files": self._modified_files,
            "total_size": self.total_size,
            "total_size_compressed": self.total_size_compressed,
            "file_count": self.file_count,
            "status": self._status,
        }
        self._upload_state()
        self._save(
            method="update",
            data=data_to_be_updated,
        )
        if verbose:
            logger.info("FILES REMOVED SUCCESSFULLY")
        return self.removed_files

    @typechecked
    def _update_file_entries_post_processing(
        self,
        verbose: bool,
        removed_operation: Optional[bool] = True,
    ) -> None:
        """
        Update dataset state
           the dataset's state changes every time we call add_files or remove_files
        :param verbose: bool: If True, enable console output for progress information
        :param removed_operation: bool: In case post-processing is triggered following a removal operation.
        """
        self._update_dependency_graph()
        self._file_count = len(self.file_entries)

        parent_file_entries: FileEntries = self._get_parent_file_entries()
        added_files: int = 0
        modified_files: int = 0
        total_size: int = 0

        for file in tqdm(
            self._file_entries.values(),
            desc="processing",
            unit=" file",
            disable=not verbose,
        ):
            total_size += file.size
            if file.parent_dataset_id == self._id:
                if file.relative_path in parent_file_entries:
                    modified_files += 1
                else:
                    added_files += 1

        self._added_files = added_files
        self._modified_files = modified_files
        self._total_size = total_size
        if removed_operation:
            removed_files: int = 0
            for parent_entry_key, parent_entry_value in tqdm(
                parent_file_entries.items(),
                desc="removal processing",
                unit=" file",
                disable=removed_operation,
            ):
                if parent_entry_key not in self._file_entries:
                    removed_files += 1
            self._removed_files = removed_files

    def _build_dependency_chunk_lookup(self) -> Dict[str, int]:
        """
        Build the dependency dataset id to number-of-chunks
        :return: dict[str,int] lookup dictionary from dataset-id to number of chunks
        """
        chunks_lookup: map = map(
            lambda dataset_id: (
                dataset_id,
                (
                    self if dataset_id == self.id else DatasetClient.get(dataset_id)
                )._get_num_chunks(),
            ),
            self._dependency_graph.keys(),
        )
        return dict(chunks_lookup)

    def _get_num_chunks(self) -> int:
        """
        Return the number of chunks stored on this dataset
           it does not imply on the number of chunks parent versions store
        :return: int: Number of chunks stored on the dataset.
        """
        artifacts: List[str] = self._get_data_artifact_names()

        return len(artifacts)

    def _get_data_artifact_names(self) -> List[str]:
        """
        Get all the artifact names in the dataset including parent artifacts
        :return: list[str]: return all the artifact names
        """
        artifacts: List[str] = []
        for file in self.file_entries:
            if (
                file.parent_dataset_id == self.id
                and file.artifact_name not in artifacts
            ):
                artifacts.append(file.artifact_name)

        return artifacts

    def _get_artifacts_to_download(self) -> Dict[str, List[Dict]]:
        """
        Get the necessary information to download all files in the dataset and its parent datasets.
            we need artifact_name and remote_url_basename(file name in the could storage)
        :return: dict[str,list[dict]] example
            {dataset_id:[{artifact_name:"..",remote_url_basename:".."},..]},
            {parent_1_id:[]},
            ...}}
        """
        artifacts: Dict[str, List] = {}
        # to keep trace of the added artifacts
        already_added: Dict = {}
        for file in self.file_entries:
            if (
                file.parent_dataset_id in artifacts
                and file.artifact_name not in already_added[file.parent_dataset_id]
            ):
                artifacts[file.parent_dataset_id].append(
                    {
                        "artifact_name": file.artifact_name,
                        "remote_url_basename": file.remote_url_basename,
                    }
                )
                already_added[file.parent_dataset_id].append(file.artifact_name)
            elif file.parent_dataset_id not in artifacts:
                artifacts[file.parent_dataset_id] = [
                    {
                        "artifact_name": file.artifact_name,
                        "remote_url_basename": file.remote_url_basename,
                    }
                ]
                already_added[file.parent_dataset_id] = [file.artifact_name]

        return artifacts

    def _get_parent_file_entries(self) -> FileEntries:
        """
        Get parent file entries
        :return: dict[str,file_entry]: example
            {parent_1_id:parent_1_file_entries},
            ...}}
        """
        parent_datasets_ids: List[str] = self._dependency_graph[self._id]
        parent_file_entries: FileEntries = dict()
        for parent_dataset_id in parent_datasets_ids:
            parent_dataset = self.get(parent_dataset_id)
            parent_file_entries.update(parent_dataset._file_entries)

        return parent_file_entries

    def _update_dependency_graph(self) -> None:
        """
        Update the dependency graph based on the current self._file_entries state
        :return: none
        """
        # collect all dataset versions
        used_dataset_versions: Set = set(
            f.parent_dataset_id for f in self._file_entries.values()
        )
        used_dataset_versions.add(self._id)
        current_parents: List = self._dependency_graph.get(self._id) or []
        # remove parent versions we no longer need from the main version list
        # per version, remove unnecessary parent versions, if we do not need them
        self._dependency_graph = {
            k: [p for p in parents or [] if p in used_dataset_versions]
            for k, parents in self._dependency_graph.items()
            if k in used_dataset_versions
        }
        # make sure we do not remove our parents
        self._dependency_graph[self._id] = current_parents

    @typechecked
    def _upload_state(self) -> None:
        """
        Upload the state of the dataset to the cloud storage service.
        """
        logger.debug("uploading state")
        state: DatasetState = dict(
            file_count=self._file_count,
            total_size=self._total_size,
            file_entries=[f.serializer() for f in self._file_entries.values()],
            dependency_graph=self._dependency_graph,
            dependency_chunk_lookup=self._dependency_chunk_lookup,
            id=self._id,
        )
        object_name: str = self.generate_state_object_name(self._name, self._id)
        StorageManager.upload_state(object_name=object_name, dataset_state=state)
        logger.debug("state uploaded")

    def _download_state(self) -> DatasetState:
        """
        Download the state of the dataset to the cloud storage service.
        :return: DatasetState: a dict that represent the sate of the dataset.
        """
        object_name: str = self.generate_state_object_name(self._name, self._id)
        state: DatasetState = StorageManager.download_state(object_name)
        return state

    @check_connection
    @typechecked
    def upload(
        self,
        compression: Optional[int] = None,
        chunk_size: Optional[int] = None,
        max_workers: Optional[int] = None,
        verbose: Optional[bool] = None,
    ) -> None:
        """
        Start file uploading, the function returns when all files are uploaded.

        :param compression: int: Compression algorithm for the Zipped dataset file (default: ZIP_DEFLATED)
        :param chunk_size: int: Artifact chunk size (MB) for the compressed dataset,
            if not provided (None) use the default chunk size (512mb).
        :param max_workers: int:Numbers of threads to be spawned when zipping and uploading the files.
            If None number of logical cores
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: none
        """
        if self.is_final():
            raise Exception(
                "Files can't be uploaded as the dataset has already been finalized."
            )
        if verbose is None:
            verbose = self.verbose

        if verbose:
            logger.info("UPLOADING FILES ...")
        if not max_workers:
            max_workers = psutil.cpu_count()

        total_size: int = 0
        chunks_count: int = 0
        keep_as_file_entry: Set = set()
        chunk_size = int(self._dataset_chunk_size_mb if not chunk_size else chunk_size)
        upload_tasks: List = []

        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            parallel_zipper = ParallelZipper(
                chunk_size,
                max_workers,
                allow_zip_64=True,
                compression=ZIP_DEFLATED if compression is None else compression,
                zip_prefix="dataset.{}.".format(self._id),
                zip_suffix=".zip",
                pool=pool,
            )
            file_paths: List[str] = []
            zip_names: Dict[str, str] = {}
            # expected number of chunks to upload
            # total_chunks value is an approximate estimation it may be lower
            total_chunks: int = self.total_size // (chunk_size * 1024 * 1024) + 1
            if verbose:
                upload_prog_bar = tqdm(
                    total=total_chunks,
                    desc="uploading",
                    unit=" chunk",
                )
            for f in self._file_entries.values():
                if not f.local_path:
                    keep_as_file_entry.add(f.relative_path)
                    continue
                file_paths.append(f.local_path)
                zip_names[f.local_path] = f.relative_path
            for zip_ in parallel_zipper.zip_iter(
                file_paths, zip_names, total_chunks, verbose
            ):
                running_tasks: List = []
                for upload_future in upload_tasks:
                    if upload_future.running():
                        running_tasks.append(upload_future)
                    else:
                        if not upload_future.result():
                            self._save(
                                "update",
                                {
                                    "id": self._id,
                                    "status": DatasetStatusEnum.failed.value,
                                },
                            )
                            raise Exception(
                                "Failed uploading dataset with ID {}".format(self._id)
                            )
                upload_tasks = running_tasks

                artifact_name: str = self._data_artifact_name
                self._data_artifact_name = self._get_next_data_artifact_name(
                    self._data_artifact_name
                )
                total_size += zip_.size
                zip_file_name: str = os.path.basename(zip_.zip_path)
                object_name: str = self.generate_object_name(
                    self._name, self._id, artifact_name, zip_file_name
                )
                upload_task = pool.submit(
                    StorageManager.upload,
                    object_name=object_name,
                    path=str(zip_.zip_path),
                )
                if verbose:
                    upload_task.add_done_callback(
                        lambda task: upload_prog_bar.update(1)
                    )
                upload_tasks.append(upload_task)
                chunks_count += 1
                for file_entry in self._file_entries.values():
                    if (
                        file_entry.local_path is not None
                        and Path(file_entry.local_path).as_posix() in zip_.files_zipped
                    ):
                        keep_as_file_entry.add(file_entry.relative_path)
                        file_entry.artifact_name = artifact_name
                        file_entry.remote_url_basename = zip_file_name
                        if file_entry.parent_dataset_id == self._id:
                            file_entry.local_path = None
                self._upload_state()

        # update the upload_prog_bar with real value of total chunks
        if verbose:
            upload_prog_bar.total = upload_prog_bar.n

        self._update_status(new_status=DatasetStatusEnum.uploaded.value)

        if chunks_count == 0:
            # No pending files, skipping upload
            data_to_be_updated: DatasetDict = {"id": self._id, "status": self._status}
            self._save("update", data_to_be_updated)

        # remove files that could not be zipped
        self._file_entries = {
            k: v
            for k, v in self._file_entries.items()
            if v.relative_path in keep_as_file_entry
        }
        self._update_file_entries_post_processing(verbose=verbose)
        data_to_be_updated = {
            "id": self._id,
            "added_files": self._added_files,
            "removed_files": self._removed_files,
            "modified_files": self._modified_files,
            "total_size": self.total_size,
            "total_size_compressed": self.total_size_compressed,
            "file_count": self.file_count,
            "status": self._status,
        }
        self._save("update", data_to_be_updated)
        self._upload_state()
        if verbose:
            logger.info("FILES UPLOADED SUCCESSFULLY")

    @check_connection
    @typechecked
    def finalize(
        self,
        auto_upload=False,
        verbose: Optional[bool] = None,
    ) -> None:
        """
        Finalize the dataset. Upload must first be called to verify that there are no pending uploads.
        If files do need to be uploaded, it throws an exception.
        :param auto_upload: bool: Automatically upload dataset if not called yet, will upload to default location.
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        """
        # check we do not have files waiting for upload.

        if verbose is None:
            verbose = self.verbose
        if self._status == DatasetStatusEnum.in_progress.value:
            if auto_upload:
                self.upload()
            else:
                self._save(
                    "update", {"id": self._id, "status": DatasetStatusEnum.failed.value}
                )
                raise Exception(
                    "can't finalize dataset, pending uploads. Call Dataset.upload(...)"
                )

        self._update_status(new_status=DatasetStatusEnum.completed.value)
        self._dependency_chunk_lookup = self._build_dependency_chunk_lookup()

        self._upload_state()
        data_to_be_updated: DatasetDict = {"id": self._id, "status": self._status}
        self._save("update", data_to_be_updated)
        if verbose:
            logger.info("DATASET COMPLETED SUCCESSFULLY")

    @classmethod
    @typechecked
    def generate_object_name(
        cls, dataset_name: str, dataset_id: str, artifact_name: str, file_path: str
    ) -> str:
        """
        Generate object name (object that we need to upload to the cloud storage)
        :param dataset_name: str: name of the dataset
        :param dataset_id: str: id of the dataset
        :param artifact_name:: str: artifact name
        :param file_path: str: file path
        :return: str: object name
        """
        workspace_name: str = os.environ["OIP_WORKSPACE_NAME"]
        object_name: str = os.path.join(
            cls.__dataset_entity_name,
            workspace_name,
            cls.__datasets_entry_name,
            dataset_name,
            f"{dataset_name}.{dataset_id}",
            artifact_name,
            file_path,
        )
        return object_name

    @classmethod
    @typechecked
    def generate_state_object_name(cls, dataset_name: str, dataset_id: str) -> str:
        """
        Generate state object name (state object that we need to upload to the cloud storage)
        :param dataset_name: str: name of the dataset
        :param dataset_id: str: id of the dataset
        :return: str: state object name
        """
        workspace_name = os.environ["OIP_WORKSPACE_NAME"]
        object_name = os.path.join(
            cls.__dataset_entity_name,
            workspace_name,
            cls.__datasets_entry_name,
            dataset_name,
            f"{dataset_name}.{dataset_id}",
            cls.__state_entry_name,
            f"{cls.__state_entry_name}.json",
        )
        return object_name

    def serializer(self) -> DatasetDict:
        """
        Serializes the object data into dict
        :return:Dataset: return a dict that represent the dataset
        """

        file_entries: List[FileEntryDict] = [f.serializer() for f in self.file_entries]

        return {
            "id": self.id,
            "name": self._name,
            "status": self._status,
            "version": self._version,
            "is_main": self._is_main,
            "parent_datasets": self._parent_datasets,
            "_tags": self._tags,
            "workspace_id": self._workspace_id,
            "description": self._description,
            "file_count": self.file_count,
            "added_files": self._added_files,
            "removed_files": self._removed_files,
            "modified_files": self._modified_files,
            "total_size": self.total_size,
            "total_size_compressed": self.total_size_compressed,
            "file_entries": file_entries,
            "dependency_graph": self._dependency_graph,
            "dependency_chunk_lookup": self._dependency_chunk_lookup,
        }

    def abort(self) -> None:
        """
        Abort a dataset
        :return:none
        """
        if self.is_final():
            raise Exception(
                "You can't abort a dataset that has already been finalized."
            )
        self._update_status(new_status=DatasetStatusEnum.aborted.value)
        data_to_be_updated: DatasetDict = {"id": self._id, "status": self._status}
        self._upload_state()
        self._save(method="update", data=data_to_be_updated)
        logger.info("DATASET ABORTED SUCCESSFULLY")

    @classmethod
    @typechecked
    def _save(cls, method: str, data: DatasetDict) -> None:
        """
        Save the dataset on the server
        :param method: str: "Create" a new dataset if it doesn't exist, or "update" an existing one .
        :param data: dict: Data we need to create or update.
        :return: none
        """
        logger.debug("call _save")
        ServerFacade.save_data(
            method,
            data,
            cls.__dataset_entity_name,
        )

    @staticmethod
    @typechecked
    def _calc_file_hash(file_entry: FileEntry):
        """
        Calculate file hash and add it to the file entry
        :param file_entry:FileEntry:
        :return: FineEntry
        """
        file_entry.hash, _ = sha256sum(file_entry.local_path)
        file_entry.size = Path(file_entry.local_path).stat().st_size
        return file_entry

    def is_completed(self) -> bool:
        """
        Return True if the dataset was completed and can't be changed any more.
        :return: bool: True if dataset is completed correctly
        """
        return self._status == DatasetStatusEnum.completed

    def is_final(self) -> bool:
        """
        Return True if the dataset was finalized and can't be changed any more.
        :return: bool: True if the dataset has reached a final status, which can be one of the following:
            completed, aborted, or failed
        """
        return self._status in [
            DatasetStatusEnum.completed.value,
            DatasetStatusEnum.aborted.value,
            DatasetStatusEnum.failed.value,
        ]

    @classmethod
    @typechecked
    def migrate(
        cls,
        storage_conf_name: str,
        download_uri: str,
        name: str,
        parent_datasets: Optional[List[str]] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        verbose: Optional[bool] = None,
        max_workers: Optional[int] = None,
    ) -> None:
        """
        Migrate dataset files for remote_uri to oip cloud storage
        :param download_uri: str: uri from where we download the dataset files refers to the specific location or path
            within the storage provider where the dataset files are hosted.
            The following are examples of ``download_uri``
             - S3: s3://bucket/folder/../dataset_folder
             - Google Cloud Storage: gs://bucket/folder/../dataset_folder
             - Azure Storage: azure://bucket/folder/../dataset_folder
             - Minio Storage: minio://bucket/folder/../dataset_folder
             - Other https://path_to_the_dataset_folder
        :param name: str: Naming the new dataset.
        :param version: str: version of the new dataset. If not set, try to find the latest version
        :param parent_datasets: list[str]: A list of parent datasets to extend the new dataset
            by adding all the files from their respective parent datasets
        :param description: str: Description of the dataset
        :param parent_datasets: list[str]: Descriptive tags categorize datasets by keywords, detailing their
            subject matter, domain, or specific topics for better identification and organization.
        :param tags: List[str]: List of tags to be associated with the dataset
        :param verbose: bool: If True, enable console output for progress information (defaults to true)
        :return: none
        """
        logger.debug("call dataset migrate")

        check_storage_conf(storage_conf_name)

        if not verbose:
            verbose = cls.verbose

        if verbose:
            logger.info("Migrating to OIP storage...")
        with tempfile.TemporaryDirectory() as tmp_dirname:
            max_workers = max_workers or psutil.cpu_count()
            dataset_path: str = StorageManager.download_via_storage_provider(
                storage_conf_name=storage_conf_name,
                download_uri=download_uri,
                download_folder=tmp_dirname,
                max_workers=max_workers,
                verbose=verbose,
            )

            new_dataset = cls.create(
                name=name,
                parent_datasets=parent_datasets,
                version=version,
                description=description,
                tags=tags,
                verbose=verbose,
            )
            new_dataset.add_files(dataset_path)
            new_dataset.upload()

            new_dataset.finalize()
        if verbose:
            logger.info(
                f"Migration completed successfully. A new dataset named {name} has been added to the OIP storage."
            )



import os
from ..dataset import DatasetClient
from ..StorageManager import StorageManager


api_host = "http://192.168.1.35:8002"
workspace_name = "default_workspace"
api_key = "4fc086fe-d64b-4ac8-816a-0dcbf34f7368"
workspace_id = "8fb7c66a-a49d-46a8-8044-51a94cee1a36"
DatasetClient.connect(api_host=api_host, api_key=api_key, workspace_name=workspace_name)
assert os.environ["OIP_API_HOST"] == api_host
assert os.environ["WORKSPACE_ID"] == workspace_id
assert os.environ["ML_DATASET_AUTHENTICATION_TOKEN"] == api_key

cifar_path = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
StorageManager.get_local_copy(remote_url=cifar_path)
